#ifndef _VFF_FILE_H_
#define _VFF_FILE_H_

#include <velib/base/types.h>
#include <stdio.h>

#define VFF_FILE_ID_STR				"Victron Energy firmware file"
#define VFF_FILE_VERSION			5

#define VFF_N_CODE_PARTS			4 /*!< Maximum number of supported code parts (Main, Bootblock, ...) */

/* Error definitions */
#define VFF_ERR_MALLOC				40 /*!< Malloc error */

#define VFF_ERR_NONE				0
#define VFF_ERR_FILE_OPEN			50 /*!< File open error */
#define VFF_ERR_FILE_WRITE			51 /*!< File write error */
#define VFF_ERR_FILE_READ			52 /*!< File read error */
#define VFF_ERR_FILE_CHECKSUM		53 /*!< File checksum error */
#define VFF_ERR_FILE_VERSION		54 /*!< File incompatible version number */
#define VFF_ERR_FILE_SECTION		55 /*!< File section not found */
#define VFF_ERR_FILE_FORMAT			56 /*!< File format error */

/* Code part names  */
#define CPN_MAIN				"Main"
#define CPN_REBOOTER			"Bootblock"
#define CPN_BOOTLOADER			"Bootloader"
#define CPN_SPECIAL_BLOCKS		"SpecialBlocks"
#define CPN_VARIABLE_LIST		"VariableList"

/* Code part tags */
#define CPT_PART						"Part"
#define CPT_EEPROM						"EL"
#define CPT_CODE						"CL"
#define CPT_MESSAGE						"ML"
#define CPT_MIN_BOOTBLOCK_FUNC			"MinBFV"
#define CPT_MUSTBE_BOOTBLOCK_FUNC		"MustBeBFV"
#define CPT_NONCE						"Nonce"
#define CPT_USED_KEY					"UsedKey"
#define CPT_KEY_ID						"KeyID"	/* ID to check if the correct family key is used. */
#define CPT_CONNECT_INFO				"ConnectInfo"
#define CPT_DEVICE						"Device"
#define CPT_VERSION						"Version"

typedef struct {
	FILE* pFile;
	int endLocation;
	un32 checksum;					/* Temp location used for on the fly checksum calculation */
} VffFile;


typedef struct {
	/*
	 * Parameters(TStringList)  with info
	 *
	 * Lines:StringList containing
	 *       - calibration map   // 1 line!!!! no info in parameters
	 *       -  Messages
	 *       - EEPROM strings
	 *       - Program strings
	 *       The NrOf lines for each of these parts is specified by the parameters
	 *
	 * ParameterLines,CalibrationMap, EEPROM and Code lines are Delphi type strings with 2 heading length bytes
	 * Message lines are C type strings with '\n' separated lines. No length bytes.
	 */
	un16 fileVersion;			/* we can have multiple files(codeparts) in one file */
	char* pFileFormatId;
	char* pParameterLines;		/* Each line starts with an un16 line length */
	char* pCalibrationMap;
	char* pMessageLines;
	char* pEEPROMLines;			/* Each line contains HID(2bytes), TID(4byte), data(64bytes), location(3bytes) */
	char* pCodeLines;
} VffCodePart;

#define VFF_EEPROM_LINE_LENGTH 2 + 4 + 64 + 3

typedef struct {
	char* pStart;
	char* pWrite;
	char* pRead;
	un32 size;
	int nCodeParts;
	VffCodePart codePart[VFF_N_CODE_PARTS];
} VffData;

void vffFileInit (FILE* pFile, VffFile* pVffFile);
int vffFileOpen (VffFile* pVffFile, char* fileName);
void vffFileClose (VffFile* pVffFile);
int vffMallocDataBuffer(VffData* buf, int size);
void vffFreeDataBuffer(VffData* buf);
int vffParseFile(VffFile* pVffFile, VffData* pData);
int vffGetIndexOfCodePart(VffData* pData, char* pName);
int vffGetParameterValueAsString(VffData* pData, int cpIndex, char* pParValue, int parValLen, char* pName);
veBool vffGetParameterValueAsInt(VffData* pData, int cpIndex, int* pValue, char* pName);
char* vffGetLineLength (un16* pVal, char* rp);
void vffEepromLineBufToRaw(char* eepromLines, char* eepromRaw);
void vffMergeEepromRawInLineBuf(char* eepromRaw, char* eepromLines);
void vffDataBufferToLog(VffData* pData);

#endif	/* _VFF_FILE_H_ */
