#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <velib/base/ve_string.h>
#include <velib/mk2/mk2_info.h>
#include <velib/mk2/vebus_device.h>

/**
 * The settings and ram scaling must be obtained from the vebus device,
 * which takes some time. To speed up initialisation they might be stored.
 * Depending on the device type this can be done in flash / EEPROM etc.
 * This is the filesystem version as used by pc's.
 */

static char *cacheDir = "cache/";

static void ramFileName(char *buf, size_t bufLen, un32 version)
{
	ve_snprintf(buf, bufLen, "%s%d.ram", cacheDir, version);
}

static FILE* openRamFile(un32 version, char const* mode)
{
	char buf[PATH_MAX + NAME_MAX + 2];
	ramFileName(buf, sizeof(buf), version);
	return fopen(buf, mode);
}

void mk2InfoRamStore(un32 version, VebusRamInfo* info, un16 numVars)
{
	FILE *fh;
	char buf[PATH_MAX + NAME_MAX + 2];

	ramFileName(buf, sizeof(buf), version);

	if (!(fh = fopen("tmp", "wb")))
		return;

	fwrite(info, sizeof(VebusRamInfo), numVars, fh);
	fclose(fh);

	sync();
	rename("tmp", buf);
	sync();
}

veBool mk2InfoRamLoad(un32 version,  VebusRamInfo* info, un16 numVars)
{
	FILE *fh;

	if (!(fh = openRamFile(version, "rb")))
		return veFalse;

	if (fread(info, sizeof(VebusRamInfo), numVars, fh) < numVars) {
		fclose(fh);
		return veFalse;
	}

	fclose(fh);
	return veTrue;
}

static void settingsFileName(char *buf, size_t bufLen, un32 version)
{
	ve_snprintf(buf, bufLen, "%s%d.set", cacheDir, version);
}

static FILE* openSettingsFile(un32 version, char const* mode)
{
	char buf[PATH_MAX + NAME_MAX + 2];
	settingsFileName(buf, sizeof(buf), version);
	return fopen(buf, mode);
}

void mk2InfoSettingsStore(un32 version, VebusSettingInfo* info, un16 numSettings)
{
	FILE *fh;
	char buf[PATH_MAX + NAME_MAX + 2];

	settingsFileName(buf, sizeof(buf), version);

	if (!(fh = fopen("tmp", "wb")))
		return;

	fwrite(info, sizeof(VebusSettingInfo), numSettings, fh);
	fclose(fh);

	sync();
	rename("tmp", buf);
	sync();
}

veBool mk2InfoSettingsLoad(un32 version, VebusSettingInfo* info, un16 numSettings)
{
	FILE *fh;

	if (!(fh = openSettingsFile(version, "rb")))
		return veFalse;

	if (fread(info, sizeof(VebusSettingInfo), numSettings, fh) < numSettings) {
		fclose(fh);
		return veFalse;
	}

	fclose(fh);
	return veTrue;
}

veBool mk2InfoClearCache(void)
{
	DIR *dir;
	struct dirent *ent;

	dir = opendir(cacheDir);
	if (dir == NULL)
		return veFalse;

	while ((ent = readdir(dir)) != NULL) {
		char fullpath[PATH_MAX + NAME_MAX + 2];
		if ((ent->d_type & DT_REG) == 0)
			continue;
		snprintf(fullpath, sizeof(fullpath), "%s%s", cacheDir, ent->d_name);
		remove(fullpath);
	}
	closedir(dir);

	return veTrue;
}
