/**
 * The items is a tree of the available values related to the vebus system and
 * the devices in it. This is optional, the intention is to provide a common
 * access to the values and hide the implementation details of the mk2 / vebus.
 */

#include <stdlib.h>

#include <velib/velib_inc_mk2.h>
#include <velib/base/ve_string.h>
#include <velib/mk2/vebus_items.h>
#include <velib/mk2/vebus_values.h>
#include <velib/mk2/mk2_strings.h>
#include <velib/utils/ve_assert.h>
#include <velib/types/ve_item.h>
#include <velib/types/ve_str.h>
#include <velib/utils/ve_item_utils.h>
#include <velib/utils/ve_logger.h>

VebusDevice *vebusItemGetDev(VeItem *item)
{
	VeBusLookup *itemCtx;
	veAssert(item);

	/* If the item is not associate with a vebus variable at all, bail out */
	itemCtx = veItemCtx(item)->ptr;
	if (itemCtx == NULL)
		return NULL;

	/* regular device, so return it */
	if (itemCtx->dev != NULL)
		return itemCtx->dev;

	/* system variables, do need scaling, so use the master for that */
	return vebusSystem.master;
}

un16 vebusItemGetVarNr(VeItem *item)
{
	VeBusLookup *itemCtx = veItemCtx(item)->ptr;
	if (!itemCtx)
		return VEBUS_ID_INVALID;
	return itemCtx->nr;
}

/* textual representation of a vebus value */
static size_t textVersion(VeVariant *var, void const *ctx, char *buf, size_t len)
{
	VeItem *item = (VeItem *) ctx;
	un16 nr = vebusItemGetVarNr(item);

	if (veVariantIsValid(var)) {
		VebusUnit unit = vebusGetUnit(nr, VE_UNITS_VSC);
		if (unit == VE_U_PERC_FRAC) {
			VeVariant perc = *var;
			perc.value.Float /= 100;
			return vebusFormatValue(unit, &perc, buf, len);
		}

		return vebusFormatValue(unit, var, buf, len);
	}

	return ve_snprintf(buf, len, "%s", "");
}

static size_t vebusStateString(VeVariant *var, void const *ctx, char *buf, size_t len)
{
	VE_UNUSED(ctx);

	if (!veVariantIsValid(var)) {
		buf[0] = 0;
		return 0;
	}

	return vebusStateText(var->value.UN32, buf, len);
}

static size_t vebusSubStateString(VeVariant *var, void const *ctx, char *buf, size_t len)
{
	VE_UNUSED(ctx);

	if (!veVariantIsValid(var)) {
		buf[0] = 0;
		return 0;
	}

	return vebusSubStateText(var->value.UN32, buf, len);
}

static size_t vebusModeString(VeVariant *var, void const *ctx, char *buf, size_t len)
{
	VE_UNUSED(ctx);

	if (!veVariantIsValid(var)) {
		buf[0] = 0;
		return 0;
	}

	return vebusModeText(var->value.UN32, buf, len);
}

static size_t vebusLedString(VeVariant *var, void const *ctx, char *buf, size_t len)
{
	VE_UNUSED(ctx);

	if (!veVariantIsValid(var)) {
		buf[0] = 0;
		return 0;
	}

	return vebusLedStateText(var->value.UN32, buf, len);
}

size_t getDescription(struct VeItem *item, char *buf, size_t len)
{
	return ve_snprintf(buf, len, "%s", vebusGetName(veItemCtx(item)->value));
}

static void vebusItemAddCtx(VeItem *item, VebusDevice *dev, un16 nr)
{
	VeBusLookup *itemCtx;

	itemCtx = veItemCtx(item)->ptr;
	if (itemCtx == 0) {			/* calloc is used to allocate mem for item
								 * (NULL is not guaranteed to be 0) */
		itemCtx = malloc(sizeof(VeBusLookup));
		if (itemCtx == NULL) {
			logE("vebus_items", "out of memory!!!");
			exit(1);
		}
	}

	itemCtx->dev = dev;
	itemCtx->nr = nr;
	veItemCtx(item)->ptr = itemCtx;
}

/* create a item corresponding to a vebus value */
static void addItem(VeItem *root, VebusDevice *dev, un16 nr, char const *uid, VeItem *item)
{
	VeVariant v;

	vebusItemAddCtx(item, dev, nr);
	veItemAddChildByUid(root, uid, item);
	veItemSetFmt(item, textVersion, item);
	veItemSetGetDescr(item, getDescription);
	v.type.tp = VE_FLOAT;
	veVariantInvalidate(&v);
	veItemOwnerSet(item, &v);
}

/* Creates / allocates a new item with a unit and attach vebus info to it */
struct VeItem *vebusItemCreateQuantity(struct VeItem *parent, VebusDevice *dev, un16 nr, char const *uid, VeVariant *initialValue, VeVariantUnitFmt *unit)
{
	struct VeItem *item;

	item = veItemGetOrCreateUid(parent, uid);
	veItemSetFmt(item, veVariantFmt, unit);
	vebusItemAddCtx(item, dev, nr);
	veItemOwnerSet(item, initialValue);

	return item;
}

#ifdef CFG_VEBUS_DEVICE_ITEMS

static veBool handleSettingChange(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VebusDevice *dev = vebusItemGetDev(item);
	un16 nr = vebusItemGetVarNr(item);
	VE_UNUSED(ctx);

	if (!dev)
		return veFalse;

	veVariantToFloat(variant);
	if (!vebusLocalSetSetting(dev, nr, variant->value.Float))
		return veFalse;

	mk2ActSetRawSetting(dev, nr);

	return veTrue;
}

static void handleGetSetting(struct VeItem *item)
{
	VebusDevice *dev = vebusItemGetDev(item);
	un16 nr = vebusItemGetVarNr(item);

	//logI("handleGetSetting", "%p %p %s %d", dev, item, veItemId(item), nr);
	if (!dev)
		return;

	mk2ActGetRawSetting(dev, nr);
}

/* create the item and also register its callbacks */
static void addSetting(struct VeItem *root, VebusDevice *dev, un16 nr, char const *uid, struct VeItem *item)
{
	addItem(root, dev, nr, uid, item);
	veItemSetSetter(item, handleSettingChange, NULL);
	veItemSetGetter(item, handleGetSetting);
}

static veBool handleFlagChange(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VebusDevice *dev = vebusItemGetDev(item);
	un16 nr = vebusItemGetVarNr(item);
	VE_UNUSED(ctx);

	if (!dev)
		return veFalse;

	veVariantToN32(variant);
	if (!vebusLocalSetFlag(dev, nr, variant->value.UN32 != 0))
		return veFalse;

	mk2ActSetFlag(dev, nr);

	return veTrue;
}

static void handleGetFlag(struct VeItem *item)
{
	VebusDevice *dev = vebusItemGetDev(item);
	un16 nr = vebusItemGetVarNr(item);

	if (!dev)
		return;

	mk2ActGetRawFlag(dev, nr);
}

/* create the item and also register its callbacks */
static void addFlag(VeItem* root, VebusDevice *dev, un16 nr, char const *uid, VeItem *item)
{
	addItem(root, dev, nr, uid, item);
	veItemSetSetter(item, handleFlagChange, NULL);
	veItemSetGetter(item, handleGetFlag);
}

/*
 * Create the device specific values.
 *
 * @param root The tree to add the device to..
 * @param sys Items of the vebusSystem
 * @param dev The device to add
 */
void vebusItemsCreateDevice(VeItem *root, VeBusItems *sys, VebusDevice *dev)
{
	VeStr s;
	VeItem *settings;
	un8 n;

	veStrNewFormat(&s, "Devices/%d/Settings", dev->addr);
	settings = veItemGetOrCreateUid(root, veStrCStr(&s));
	veStrFree(&s);

#define X(a,b)	addSetting(settings, dev, a, b, &sys->devices[dev->addr].settings[a]);
	VE_DEVICE_SETTINGS
#undef X

#define X(a,b)	addFlag(settings, dev, a, b, &sys->devices[dev->addr].flags[(un8) a]);
	VE_DEVICE_FLAGS
#undef X

	for (n = 0; n < VEBUS_SETTINGS_COUNT; n++) {
		VeVariant var;

		veVariantFloat(&var, vebusGetSettingMaximum(dev, n, VE_UNITS_VSC));
		veItemSetMax(&sys->devices[dev->addr].settings[n], &var);

		veVariantFloat(&var, vebusGetSettingMinimum(dev, n, VE_UNITS_VSC));
		veItemSetMin(&sys->devices[dev->addr].settings[n], &var);
	}
}

#else

void vebusItemsCreateDevice(VeItem *root, VeBusItems *sys)
{
	VE_UNUSED(root);
	VE_UNUSED(sys);
}

#endif


/* Performs unit conversion and creates on change event for the item tree */
void vebusItemUpdate(VeItem *item)
{
	VeVariant value;
	VebusDevice *dev = vebusItemGetDev(item);
	un16 nr = vebusItemGetVarNr(item);

	if (!dev) {
		logW("vebus_items", "no device associated with item, refusing to update!");
		return;
	}
	// logI("updateItem", "%s %d %d", veItemId(item), dev->addr, nr);
	vebusGetValue(dev, nr, &value, VE_UNITS_VSC);

	/* [0..1] to [0..100] range conversion */
	if (nr == VE_VAR_SOC && veVariantIsValid(&value))
		value.value.Float *= 100;

	veItemOwnerSet(item, &value);
}

/**
 * Create an item tree with vebus values.
 *
 * The tree with items can be used to navigate through the values in
 * vebus system, format / set / get them etc.
 */
void vebusItemsCreateSystem(VeItem *root, VeBusItems *sys)
{
#define X(a,b)	addItem(root, NULL, a, b, &sys->a);
	VE_SYSTEM_CREATE_ITEMS
#undef X

	vebusItemUpdate(&sys->VE_SYS_DMC_VERSION);
	vebusItemUpdate(&sys->VE_SYS_BMS_VERSION);

	veItemSetFmt(&sys->VE_SYS_MAINSTATE, vebusStateString, NULL);
	veItemSetFmt(&sys->VE_SYS_SUBSTATE, vebusSubStateString, NULL);
	veItemSetFmt(&sys->VE_SYS_MODE, vebusModeString, NULL);

	veItemSetFmt(&sys->VE_SYS_LED_MAINS, vebusLedString, NULL);
	veItemSetFmt(&sys->VE_SYS_LED_BULK, vebusLedString, NULL);
	veItemSetFmt(&sys->VE_SYS_LED_ABSORPTION, vebusLedString, NULL);
	veItemSetFmt(&sys->VE_SYS_LED_FLOAT, vebusLedString, NULL);
	veItemSetFmt(&sys->VE_SYS_LED_INVERTER, vebusLedString, NULL);
	veItemSetFmt(&sys->VE_SYS_LED_OVERLOAD, vebusLedString, NULL);
	veItemSetFmt(&sys->VE_SYS_LED_LOW_BATTERY, vebusLedString, NULL);
	veItemSetFmt(&sys->VE_SYS_LED_TEMPERATURE, vebusLedString, NULL);

	veItemSetFmt(&sys->VE_SYS_ALARM_L1_LOW_BATTERY, veItemAlarmText, NULL);
	veItemSetFmt(&sys->VE_SYS_ALARM_L1_TEMPERATURE, veItemAlarmText, NULL);
	veItemSetFmt(&sys->VE_SYS_ALARM_L1_OVERLOAD, veItemAlarmText, NULL);
	veItemSetFmt(&sys->VE_SYS_ALARM_L1_U_BAT_RIPPLE, veItemAlarmText, NULL);

	veItemSetFmt(&sys->VE_SYS_ALARM_L2_LOW_BATTERY, veItemAlarmText, NULL);
	veItemSetFmt(&sys->VE_SYS_ALARM_L2_TEMPERATURE, veItemAlarmText, NULL);
	veItemSetFmt(&sys->VE_SYS_ALARM_L2_OVERLOAD, veItemAlarmText, NULL);
	veItemSetFmt(&sys->VE_SYS_ALARM_L2_U_BAT_RIPPLE, veItemAlarmText, NULL);

	veItemSetFmt(&sys->VE_SYS_ALARM_L3_LOW_BATTERY, veItemAlarmText, NULL);
	veItemSetFmt(&sys->VE_SYS_ALARM_L3_TEMPERATURE, veItemAlarmText, NULL);
	veItemSetFmt(&sys->VE_SYS_ALARM_L3_OVERLOAD, veItemAlarmText, NULL);
	veItemSetFmt(&sys->VE_SYS_ALARM_L3_U_BAT_RIPPLE, veItemAlarmText, NULL);
}

#define X(a,b)	vebusItemUpdate(&sys->a);

void vebusItemsActionDone(VeBusItems *sys, VebusDevice *dev, Mk2Actions done)
{
	VE_UNUSED(dev);

	switch (done)
	{
	case MK2_ACT_AC_CONFIG:
		{
			/*
			 * The active AC in min / max is only reported once in use.
			 * So keep a copy around.
			 */
			un8 activeIn = vebusActiveIn(&vebusSystem.acInConfig);
			if (activeIn < ARRAY_LENGTH(vebusSystem.acInConfigPerAcIn))
				vebusSystem.acInConfigPerAcIn[activeIn] = vebusSystem.acInConfig;
		}
		VE_AC_CONFIG_VALUES
		break;
	case MK2_ACT_STATUS:
		/* update the alarms derived from the leds and state */
		vebusSystem.alarms = vebusAlarmsFromState(&vebusSystem.state);
		VE_SYSTEM_ALARMS
		VE_SYSTEM_ALARMS_PER_PHASE
		VE_SYSTEM_STATE
		VE_SYSTEM_LEDS
		break;
	case MK2_ACT_CONV_STATE:
		VE_SYSTEM_CONV_VARS_VEBUS
		break;
	case MK2_ACT_DC_VALUES:
		VE_SYSTEM_DC_VARS
		break;
	case MK2_ACT_AC_VALUES:
		VE_SYSTEM_AC_ITEMS
		break;
	case MK2_ACT_GET_KWH_COUNTERS:
		VE_SYSTEM_KWH_COUNTERS
		break;
	case MK2_ACT_GET_ASSISTANT_VALUES:
		VE_SYSTEM_AC_SENSORS
		break;
	default:
		;
	}
}

#undef X

#ifdef CFG_VEBUS_DEVICE_ITEMS

/* update all flags at once */
#define X(a,b)	vebusItemUpdate(&sys->devices[dev->addr].flags[(un8) a]);

void vebusItemsSettingChanged(VeBusItems *sys, struct VebusDeviceS* dev, un8 id)
{
	logI("vebusItemsSettingChanged", "%d:%d", dev->addr, id);

	if (id == VE_FLAGS_0 || id == VE_FLAGS_1 || id == VE_FLAGS_2 || id == VE_FLAGS_3) {
		VE_DEVICE_FLAGS
		return;
	}

	vebusItemUpdate(&sys->devices[dev->addr].settings[id]);
}

#undef X

#else

void vebusItemsSettingChanged(VeBusItems *sys, struct VebusDeviceS* dev, un8 id)
{
	VE_UNUSED(sys);
	VE_UNUSED(dev);
	VE_UNUSED(id);
}

#endif

void vebusItemsRamVarObtained(VeBusItems *sys, struct VebusDeviceS* dev, un8 id)
{
	/* The SOC of the master is the SOC of the system */
	if (dev->addr == 0 && id == VE_VAR_SOC_RAMVAR)
		vebusItemUpdate(&sys->VE_VAR_SOC);

}

void vebusItemsInitDone(VeBusItems *sys, VeItem *root, VebusDevice *masterDevice)
{
	VE_UNUSED(root);
	VE_UNUSED(masterDevice);

#ifdef CFG_VEBUS_DEVICE_ITEMS

	{
		VebusDevice *dev = vebusSystem.devices;
		while (dev) {
			vebusItemsCreateDevice(root, sys, dev);
			dev = dev->next;
		}
	}
#endif

/* Set info about the vebus system */
#define X(a,b)	vebusItemUpdate(&sys->a);
	VE_SYSTEM_GENERAL
	VE_SYSTEM_STORED_DEVICE_VERSIONS
#undef X

	vebusItemUpdate(&sys->VE_SYS_MODE_IS_ADJUSTABLE);
}

#define X(a, b)	veItemInvalidate(&sys->a);
void vebusItemsInvalidate(VeBusItems *sys, VeItem *root)
{
	VE_SYSTEM_INVALIDATE_ITEMS
	VE_UNUSED(root);

	/* If the device is off, allow it to be turned on in case no BMS was detected */
	vebusItemUpdate(&sys->VE_SYS_MODE_IS_ADJUSTABLE);
}
#undef X

#if 0
/**
 * @brief request a system variable by its id
 * @param id of the system variable
 * @return wether succesfull
 */
veBool mk2ReqSystemVar(un8 id)
{
#define X(a,b)	case a:

	switch (id)
	{
	VE_SYSTEM_DC_VARS
		mk2ActMark(MK2_ACT_DC_VALUES);
		return veTrue;

	VE_SYSTEM_STATE_VARS
		mk2ActMark(MK2_ACT_STATUS);
		return veTrue;

	VE_SYSTEM_CONV_VARS
		mk2ActMark(MK2_ACT_CONV_STATE);
		return veTrue;

	VE_SYSTEM_AC_VARS
		mk2ActMark(MK2_ACT_AC_VALUES);
		return veTrue;

	VE_SYSTEM_CURRENT_LIMIT
		mk2ActMark(MK2_ACT_AC_CONFIG);
		return veTrue;

	default:
		return veFalse;
	}

#undef X
}

#endif
