/**
 * This file contains MK3 interface specific functions and should be considered as an
 * extention to the functions described mk2.c
 */
#include <ctype.h>

#include <velib/velib_inc_mk2.h>
#include <velib/base/endian.h>
#include <velib/mk2/mk2.h>
#include <velib/mk3/mk3.h>
#include <velib/utils/ve_logger.h>
#include <velib/utils/ve_assert.h>


static Mk2EventReply mk3SelfTestHandler(Mk2Event mk2event, Mk2MsgRx *msg)
{
	if (mk2event == MK2_FRAME && msg->buf[0] == 0xFF &&
			msg->buf[1] == 'T' && msg->length >= 4)
	{
		logI("mk3", "--- MK3 self test result 0x%x, 0x%x", msg->buf[2], msg->buf[3]);
		return MK2_DONE;
	}

	return MK2_CONTINUE;
}

void mk3SelfTest (Mk2Callback* callback)
{
	Mk2MsgTx* msg = mk2Msg('T');
	if (!msg)
		return;

	logI("mk3", "---- MK3 self test ----");

	msg->callback = callback;
	msg->handler = mk3SelfTestHandler;
	mk2FhQueueFrame(msg);
}

static Mk2EventReply mk3SetHardwareLinesHandler(Mk2Event mk2event, Mk2MsgRx *msg)
{
	if (mk2event == MK2_FRAME && msg->buf[0] == 0xFF &&
			msg->buf[1] == 'H' && msg->length >= 3)
	{
		logI("mk3", "--- Set panel detect and standby line levels 0x%x", msg->buf[2]);
		return MK2_DONE;
	}

	return MK2_CONTINUE;
}

/**
 *  Sets the panel detect and standby lines.
 *
 * @param callback
 * @param panel
 * @param standBy
 */
void mk3SetHardwareLines (Mk2Callback* callback, veBool panel, veBool standBy)
{
	un8 lineCtrl;

	Mk2MsgTx* msg = mk2Msg('H');
	if (!msg)
		return;

	logI("mk3", "---- Set panel detect and standby line levels ----");

	lineCtrl = 0;
	panel == veTrue ? lineCtrl |= 0x1 : lineCtrl & 0xFE; /* bit 0 is panel */
	standBy == veTrue ? lineCtrl |= 0x2 : lineCtrl & 0xFD; /* bit 1 is standby */

	msg->buf[2] = lineCtrl;
	msg->length = 3;

	msg->callback = callback;
	msg->handler = mk3SetHardwareLinesHandler;
	mk2FhQueueFrame(msg);
}


static Mk2EventReply mk3GetBootVersionsHandler(Mk2Event mk2event, Mk2MsgRx *msg)
{
	Mk3BootVersions* versions = mk2.fh.msgOut->ctx;

	if (mk2event == MK2_FRAME && msg->buf[0] == 0xFF &&
			msg->buf[1] == 'B' && msg->buf[2] == MK3_BOOTLOAD_GET_VERSIONS && msg->length >= 7)
	{
		versions->bootFunctionVersion = (un16) msg->buf[3];
		versions->bootReprogramVersion = (un16) msg->buf[4];
		versions->bootBlockVersion = le_u16_to_cpu_p(&msg->buf[5]);

		return MK2_DONE;
	}

	return MK2_CONTINUE;
}


/**
 * Get version information from the bootblock
 *
 * @param versions Pointer to BootVersions struct
 * @param callback Called when done
 */
void mk3GetBootVersions(Mk3BootVersions* versions, Mk2Callback* callback)
{
	Mk2MsgTx* msg = mk2Msg('B');
	if (!msg)
		return;

	msg->buf[2] = MK3_BOOTLOAD_GET_VERSIONS;
	msg->handler = mk3GetBootVersionsHandler;
	msg->callback = callback;
	msg->ctx = versions;
	msg->length = 3;
	logI("mk3", "---- Get bootloader version information ----");
	mk2FhQueueFrame(msg);
}


static Mk2EventReply mk3GetIDsFromBootloaderHandler(Mk2Event mk2event, Mk2MsgRx *msg)
{
	Mk3IDs* ids = mk2.fh.msgOut->ctx;

	if (mk2event == MK2_FRAME && msg->buf[0] == 0xFF &&
			msg->buf[1] == 'B' && msg->buf[2] == MK3_BOOTLOAD_GET_IDS && msg->length >= 11)
	{

		logI("mk3", "---- Got IDs from bootloader ---- ");

		ids->tid = le_u32_to_cpu_p(&msg->buf[3]);
		ids->keyId = le_u16_to_cpu_p(&msg->buf[7]);
		ids->hid = le_u16_to_cpu_p(&msg->buf[9]);
		ids->timedOut = veFalse;
		return MK2_DONE;
	}

	if (mk2event == MK2_TIMEOUT) {
		if (--ids->retriesRemaining == 0) {
			logI("mk3", "---- Get IDs timed out, no firmware? ----");
			ids->timedOut = veTrue;
			return MK2_DONE;
		}
	}

	return MK2_CONTINUE;
}

/**
 * Get the hardware and target ID info from the bootblock.
 * Use with care.
 * This function temporarily disables VEBUS communication on the MK3 in order
 * to call the appropriate bootloader function.
 *
 * MK2 communication timeout will be registered in the Mk3IDs struct.
 * This is used by the firmware update process in order to determine
 * if a vallid MK3 application is available.
 *
 * @param ids Pointer to Mk3IDs struct
 * @param callback Called when done
 * @param nRetry number of retries before giving up (should be < MK2_RETRIES)
 */
void mk3GetIDsFromBootloader(Mk3IDs* ids, Mk2Callback* callback, un8 nRetry)
{
	Mk2MsgTx* msg = mk2Msg('B');
	if (!msg)
		return;

	ids->retriesRemaining = nRetry;
	ids->timedOut = veFalse;

	msg->buf[2] = MK3_BOOTLOAD_GET_IDS;
	msg->handler = mk3GetIDsFromBootloaderHandler;
	msg->callback = callback;
	msg->ctx = ids;
	msg->length = 3;
	logI("mk3", "---- Get IDs from bootloader ----");
	mk2FhQueueFrame(msg);
}

/**
 * Call the bootloader's start reprogram function.
 * There is no reply message, MK3 firmware will be waiting in the bootloader
 * After this the firmware update routines should be called.
 */
void mk3StartReprogram(void)
{
	Mk2MsgTx* msg = mk2Msg('B');
	if (!msg)
		return;

	msg->buf[2] = MK3_BOOTLOAD_START_REPROGRAM;
	msg->handler = NULL;		/* There is no reply to this instruction */
	msg->callback = NULL;
	msg->length = 3;
	logI("mk3", "---- Send start reprogram to bootloader ----");
	mk2FhQueueFrame(msg);
}
