/**
 * This will update the firmware in the mk3 itself.
 *
 * On embedded targets some additional io lines can be used to get the mk3 in/out of
 * bootloader mode. Additional hardware functions needed to update are:
 *
 *	mk2SetBaudRate(MK2_BAUD_115200);
 *	vebusDeviceEnableUpdate(veTrue);              by-pass the normal frame handler
 *
 *	note that while vebusDeviceEnableUpdate == veTrue the incoming bytes should be passed to
 *		mk3RxUpdateRxByte
 *
 *	And 50 ms ticks must from main must call:
 *		mk3FirmwareTick
 *
 *  The MK3 bootloader PC communication starts at 2400 baud.
 *  A baudrate change to 115200 baud will be issued after retrieval of bootloader version information in order to
 *  speedup the update.
 *
 *	If Firmware update is done mk3FirmwareUpgrade(callback) will be called
 *
 *  There is a three time retry mechanism.
 *  1. Update start from valid application
 *  2. Update start from bootloader, invalid application (broken checksum)
 *  3. Update start from bootloader, forced. In case the application checksum is ok and application is not ok.
 *
 *  Update can be forced to start from 2.
 *  Add 2. With an invalid application checksum the bootloader waits in the boottest loop.
 *  Add 3. Reset of MK3 is required
 *
 *  @todo update MK3 from VEBUS? (May not be required, since we do note have "MK3"(PIC18) based BMS/digital multicontrol)
 *
 *  @note
 *  This is supported by the bootloader. In this case auto baud is used. Not sure if this a usefull feature.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <velib/mk2/mk2.h>
#include <velib/mk2/hal.h>
#include <velib/mk3/mk3.h>
#include <velib/base/endian.h>
#include <velib/utils/ve_todo.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_logger.h>
#include <velib/utils/ve_timer.h>
#include <velib/vebusupd/vbdup_misc.h>

#include <velib/vebusupd/mk3_update.h>
#include <velib/vebusupd/vebus_device_update.h>

typedef struct {
	un8 bootReprogramVersion;
	un8 bootFunctionsVersion;
	BootBlockType bootBlockType;
	un8 bootBlockRevision;
	un8 extraFlags;
} BootBlockVersionInfo;

 /* The states used during the update procedure */
typedef enum
{
	UPDATE_IDLE,
	UPDATE_START,
	UPDATE_RESETTING,
	UPDATE_SEND_ACCES_STRING,
	UPDATE_GET_IDS_THROUGH_APP,
	UPDATE_SEND_UPDATE_CMD_TO_APP,
	UPDATE_GET_VERSION,
	UPDATE_SETUP_KEYS_AND_RESPONSE_TYPE,
	UPDATE_GET_KEY_CODE,
	UPDATE_SEND_AUTOBAUD_CMD,
	UPDATE_SEND_AUTOBAUD_STRING,
	UPDATE_GET_EEPROM_VALUES,
	UPDATE_GET_CALIBRATION_MAP,
	UPDATE_SEND_EEPROM_VALUES,
	UPDATE_SEND_CODE,
	UPDATE_RUN_APPLICATION,
	UPDATE_DONE
} UpdateState;

/* Statemachine flow from application is different.
 * When the bootloader reprogram command is called from the applicaton the bootloader is locked/not selected.
 * No commands other than BBC_SETUP_KEY_AND_RESP_TYPE can be send.
 *
 * update from reset                      update from application
 * UPDATE_START                             UPDATE_START
 * UPDATE_SEND_ACCES_STRING                 UPDATE_GET_IDS_THROUGH_APP
 * UPDATE_GET_VERSION                       UPDATE_SEND_UPDATE_CMD_TO_APP
 * UPDATE_SEND_AUTOBAUD_CMD                 UPDATE_SETUP_KEYS_AND_RESPONSE_TYPE
 * UPDATE_SEND_AUTOBAUD_STRING              UPDATE_GET_VERSION
 * UPDATE_GET_KEY_CODE                      UPDATE_SEND_AUTOBAUD_CMD
 * UPDATE_SETUP_KEYS_AND_RESPONSE_TYPE      UPDATE_SEND_AUTOBAUD_STRING
 *                          Common
 *                    UPDATE_GET_EEPROM_VALUES
 *                    UPDATE_GET_CALIBRATION_MAP
 *                    UPDATE_SEND_EEPROM_VALUES
 *                    UPDATE_SEND_CODE
 *                    UPDATE_RUN_APPLICATION
 */

typedef enum
{
	UPDATE_BEGIN_WITH_RESET,
	UPDATE_BEGIN_FROM_BOOTLOADER,
	UPDATE_BEGIN_FROM_APP,
	UPDATE_TIMER,
	UPDATE_REV_VERSION,
	UPDATE_REV_NEXT_DATA,
	UPDATE_REPLY
} UpdateEvent;

typedef enum
{
	UPDATE_ERR_OK,
	UPDATE_ERR_VERSION_TIMEOUT,
	UPDATE_ERR_VERSION_INVALID,
	UPDATE_ERR_PROG_RESPONSE,
	UPDATE_ERR_BUFFER_OVERFLOW,
	UPDATE_ERR_PROG_TIMEOUT,
	UPDATE_ERR_VFF_FILE,
	UPDATE_ERR_BOOTLOADER_ACCESS_RESPONSE_TIMEOUT,
} UpdateError;

#define MK3_N_GET_ID_RETRIES	3 /* Maximum number of allowed communication retries to determine wether
								   * we are talking with a valid mk3 application */

/* Bootblock commands */
#define BBC_SETUP_KEY_AND_RESP_TYPE		0
#define BBC_PROGRAM_WRITE				1
#define BBC_EEPROM_WRITE				2
#define BBC_SEND_IDS					3
#define BBC_SEND_CALIBRATION_LOCATIONS  4
#define BBC_SEND_REPROGRAM_VERSION		5
#define BBC_MODIFY_HW_ID				6
#define BBC_RUN_RESET					7
#define BBC_EEPROM_READ					8
#define BBC_ADAPT_BAUDRATE				9

/* Bootblock responses */
#define BBR_WITHIN_BOOTBLOCK			'B'
#define BBR_CALIBRATION_MAP				'C'
#define BBR_WRONG_ID					'D'
#define BBR_ERROR						'E'
#define BBR_IDS							'I'
#define BBR_NEXT						'N'
#define BBR_VERSION						'V'
#define BBR_RANDOM_NUMBER_FOR_NONCE		'R'
#define BBR_EEPROM_READ					'O'

static char txBuffer[128];
static volatile char rxBuffer[128]; /* Contents <command>,<data bytes>,<checksum> */
static volatile un8 rxBufPos = 0;	/* This can be used ase length */
static volatile un8 bytesExpected = 0;
static volatile veBool dataExpected = veFalse;
static volatile veBool dataReady = 0;
static volatile veBool receivedStartChar = veFalse;
static volatile veBool receivedEscapeChar = veFalse;

static UpdateError updateError;			/* Last error */
static UpdateState updateState;			/* The current state of the update */
static int updateSubstateIndex;			/* Index used when states are executed multiple times */
static VffData* pVffData;				/* pointer to the parsed VFF file data */
static int vffDataCPI;					/* Index of the requested code part */
static int updateRetryCount;
static int retransmitCount;
static BootBlockVersionInfo bootBlockVersionInfo;
static char deviceCalibrationMap[32];
static char newCalibrationMap[32];
static char deviceEepromData[256];
static un32 firmwareVersionBeforeUpdate;
static un16 updateTimeout = 0;			/* in 50 ms, 0 is idle */
static VbdubUpdateResultCallback* resultCallback;
static Mk3IDs mk3IDs;
static veBool updateFromApp;

static un8 const bootloaderAccess[] = {0xC0,0xC0,0xC0,0xC0,0xC0,0xC0,0xC0,0xC0,0xC0,0xC0,0xC0,0xC0,0xC0,0xC0,0xC0,0xC0};
static un8 const autoBaudString[] = {0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55};
/* un8 const leaveTransMode[] = {0xE6, 0xE6, 0xE6, 0xE6}; */
/* The bootloader now skips transparent mode when the bootloaderAccess string is used. \
 * When using a break transparent mode is used */

static void updateResult(UpdateError error);
static void getIDsDone(void);

/**
 * Read a un32 from the rxBuffer rp and return the pointer to the next location
 *
 * @param pVal pointer to resulting value location.
 * @param rp read pointer.
 *
 * @return rp to next
 */
static char* mk3BlBufReadUn32 (un32* pVal, char* rp)
{
	* pVal = (un32)(* rp++) & 0xFF;
	* pVal |= ((un32)(* rp++) & 0xFF) << 8;
	* pVal |= ((un32)(* rp++) & 0xFF) << 16;
	* pVal |= ((un32)(* rp++) & 0xFF) << 24;

	return rp;
}

/**
 * Write a un32 to buffer wp and return the pointer to the next location
 *
 * @param val value to be written.
 * @param wp write pointer.
 *
 * @return wp to next
 */
static char* mk3BlBufWriteUn32 (un32 val, char* wp)
{
	* wp++ = (char) (val & 0xFF);
	* wp++ = (char) ((val >> 8) & 0xFF);
	* wp++ = (char) ((val >> 16) & 0xFF);
	* wp++ = (char) ((val >> 24) & 0xFF);

	return wp;
}


/**
 * Converts VFF file hex string to array of ASCII bytes.
 * Function derived from updaterunit.pas.
 *
 *
 *
 * @param pString
 * @param pBuf     a Buffer with sufficient size to contain the result.
 *                 No size checking !!!
 * @param len
 *
 * @return length of resulting buffer
 */
static int convertHexStringToASCII (char* pString, char* pBuf)
{
	int i, j, len;
	char strBuf[3];				/* intermediate buffer for 2 char strings */
	un8 result;

	len = strlen(pString);

	memset(strBuf, '\0', sizeof(strBuf)); /* Makes sure string termination is initialized */

	if (len % 2 != 0)
		return 0; /* Len should be modulo 2 */

	j = 0;
	for (i = 0; i < len; i += 2) {
		strncpy(strBuf, & pString[i], 2);
		result = (un8) strtol (strBuf, NULL, 16); /* Convert hex value */
		pBuf[j++] = result;
	}

	return j;
}

/* Command frame structure
 * 0x0, 0x8n, <command id>, <data bytes>, <Checksum>
 *
 * 0x0 => start of frame
 * 0x8n => 8 = stuffed char flag, n = data bytes excluding stuffing bytes.
 * Checksum = 1 byte. command id and data bytes. stuffing bytes are not included
 */

/**
 * Calculate checksum
 *
 * @param pData
 * @param len
 *
 * @return checksum
 */
static un8 mk3BlCalcChecksum(char* pData, int len)
{
	un8 checksum;
	int i;

	checksum = 0;
	for (i = 0; i < len; i ++)
		checksum -= (un8) pData[i];

	return checksum;
}
/**
 * Create a MK3 bootloader command frame.
 *
 * @param pDataIn Input
 * @param len DataIn buffer length
 * @param pCommandFrame Output. This buffer should be large enough. No length checking!!!
 *
 * @return pCommandFrameLength
 */
static int mk3BlCreateNewTypeCommandFrame(char* pDataIn, int len, char* pCommandFrame)
{
	int i,j;
	char dataByte;

	/* Add checksum to data */
	pDataIn[len] = mk3BlCalcChecksum(pDataIn, len); /* Checksum is excluding length */

	j = 0;
	pCommandFrame[j++] = 0x0;			/* Start of frame */
	pCommandFrame[j++] = len | 0x80;	/* Length is excluding checksum */

	len ++;								/* Include checksum in loop (incase checksum == 0) */

	for (i = 0; i < len; i ++) {
		dataByte = pDataIn[i];
		pCommandFrame[j++] = dataByte;
		if (dataByte == 0)
			pCommandFrame[j++] = 0x80; /* Byte stuffing. 0x00 is escape char */
	}

	return j;					/* command frame length */
}


/**
 * Start listening for data from the MK3
 *
 */
static void mk3BlRxStart (void)
{
	dataExpected = veFalse;
	rxBufPos = 0;
	bytesExpected = 0;
	receivedStartChar = veFalse;
	receivedEscapeChar = veFalse;
	dataReady = veFalse;
	dataExpected = veTrue;
}

static void mk3BlSendCommand(char* pComBuf, int len)
{
	int comLen;

	mk3BlRxStart();	/* Start listening for Rx data */
	comLen = mk3BlCreateNewTypeCommandFrame(pComBuf, len, txBuffer);
	mk2TxBuf((un8*)txBuffer,comLen);
}

/**
 * Check the checksum of the data in the rxBuffer
 *
 * @return veTrue if data is valid
 */
static veBool mk3BlVerifyRxData (void)
{
	un8 rxCheckSum,checkSum;

	if (rxBufPos == 0)
		return veFalse;

	rxCheckSum = (un8) rxBuffer[rxBufPos-1];

	checkSum = mk3BlCalcChecksum((char *)rxBuffer, (rxBufPos-1));

	if (checkSum == rxCheckSum)
		return veTrue;

	return veFalse;
}

/**
 * Merge (256 bytes) device EEPROM values with file EEPROM values.
 * The merged EEPROM values are written back to the VFF data structure.
 *
 * @param clearCalibration veTrue, reset calibration values to 0. skips S/N.
 *
 * @return veTrue on success
 */
static veBool mk3BlMergeCalibrationData(veBool clearCalibration)
{
	 int i;
	 un16 len;
	 char* bp;
	 char mergedEepromData[256];

	 logI("mk3_upd", "Merging file with device EEPROM data");

	 if (clearCalibration) {
		 /* Keep the S/N which is at 0xE2...0xEC */
		 memset(&newCalibrationMap, 0, sizeof(newCalibrationMap));

		 for (i = 0xE2; i < 0xEC; i ++)
			 bset(newCalibrationMap, i);
	 }

	 else {
		 /* Use the file calibration map. */
		 bp = pVffData->codePart[vffDataCPI].pCalibrationMap;
		 bp = vffGetLineLength(&len, bp);

		 /* Skip first 'C' char. according too VFFFiles.pas this is added for history reasons. */
		 len --;
		 bp ++;

		 /* Check length */
		 if (len > sizeof(newCalibrationMap)) {
			 logE("mk3_upd", "Calibration map in file exceeds device calibration map size (%d vs %d)", len, sizeof(newCalibrationMap));
			 return veFalse;
		 }

		 memcpy(newCalibrationMap, bp, len);

		 /* In case the calibration map in file is smaller, fill the remaining positions with 0xFF */
		 for (i = len; i < sizeof(newCalibrationMap); i ++)
			 newCalibrationMap[i] = 0xFF;
	 }

	 /* now we only keep those EEPROM values from device which are calibration values in BOTH!!
	  * Reason for this is that it is also possible that the new firmware has less calibration
	  * values but has settings in those locations.
	  * If we then keep the values from the device those settings will get wrong values.
	  **/

	 vffEepromLineBufToRaw(pVffData->codePart[vffDataCPI].pEEPROMLines, mergedEepromData);

	 for (i = 0; i < 256; i ++) {
		 if(btst(deviceCalibrationMap, i) && btst(newCalibrationMap, i)) {
			 /* Merge the EEPROM value from the device with the file EEPROM value. */
			 logI("mk3_upd", "Keep device EEPROM data: addr = 0x%x, data = 0x%x", i, (un8) deviceEepromData[i]);
			 mergedEepromData[i] = deviceEepromData[i];
		 }
	 }
	 vffMergeEepromRawInLineBuf(mergedEepromData,  pVffData->codePart[vffDataCPI].pEEPROMLines);

	 return veTrue;
}


/* Used twice by mk3FirmwareUpdateState.
 * The statemachine follows a slightly different path when the update starts from within the application.
 * In this case the keys should be setup first before any other command can be send.
 * When initiated from the bootloader itself, after connecting the version information is automatcally send.
 */
static void mk3BlStartGetEepromValues(char* pCommand)
{
	updateState = UPDATE_GET_EEPROM_VALUES;
	pCommand[0] = BBC_EEPROM_READ;
	pCommand[1] = updateSubstateIndex * 64;
	pCommand[2] = 0;
	mk3BlSendCommand(pCommand,3);
	updateTimeout = 5;
}


/* The update logic itself */
static void mk3FirmwareUpdateState(UpdateEvent ev)
{
	un16 temp, len;
	int value;
	un8 setupType;
	char* rp;
	char* wp;
	char command[128];
	char strBuf[128];

	switch(updateState) {
		case UPDATE_DONE:
			break;

		case UPDATE_RESETTING:
			if (ev == UPDATE_TIMER)	{
				ev = UPDATE_BEGIN_FROM_BOOTLOADER;
#if CFG_WITH_MK3_RESET
				mk3HoldInReset(veFalse);  /* mk3 release reset... */
#else
				logI("mk3_upd", "Release MK3 reset");
#endif
			} else {
				veAssert(veFalse);
				break;
			}
			/* NOTE: Fall through.*/
		case UPDATE_IDLE:
		case UPDATE_START:
			if (ev == UPDATE_BEGIN_WITH_RESET) {
#if CFG_WITH_MK3_RESET
				mk3HoldInReset(veTrue);  /* mk3 hold reset... */
#else
				logI("mk3_upd", "Please reset MK3");
#endif
				updateState = UPDATE_RESETTING;
				updateTimeout = 4;
				break;
			}
			else if (ev == UPDATE_BEGIN_FROM_BOOTLOADER || ev == UPDATE_BEGIN_FROM_APP) {
				updateError = UPDATE_ERR_OK;
				rxBufPos = 0;
				bytesExpected = 0;

				if (ev == UPDATE_BEGIN_FROM_BOOTLOADER) {
					mk2SetBaudRate(MK2_BAUD_2400); /* Default baudrate is 2400 Baud. the baudrate can be changed after sending the access string */
					updateFromApp = veFalse;
					logI("mk3_upd", "sending bootloader access string");
					updateState = UPDATE_SEND_ACCES_STRING;
					retransmitCount = 0;
					vebusDeviceEnableUpdate(veTrue); /* by-pass the normal frame handler */
					mk3BlRxStart();	/* Start listening for Rx data */
					mk2TxBuf(bootloaderAccess, sizeof(bootloaderAccess)); /* Not a standard frame */
					updateTimeout = 2;
				}
				else {
					updateFromApp = veTrue;
					logI("mk3_upd", "Get IDs from bootloader");
					updateState = UPDATE_GET_IDS_THROUGH_APP;
					mk3GetIDsFromBootloader(& mk3IDs, getIDsDone, MK3_N_GET_ID_RETRIES);
					updateTimeout = 500; /* Give the normal frame handler some time to call getIDsDone */
				}
			}
			else {
				veAssert(veFalse);
			}
			break;

		case UPDATE_GET_IDS_THROUGH_APP:
			if (ev == UPDATE_TIMER)	{
				updateResult(UPDATE_ERR_PROG_TIMEOUT);
			} else if (ev == UPDATE_REPLY) {
				/* Send update command to application */
				if(mk3IDs.timedOut) {
					logI("mk3_upd", "Start update from application failed");
					updateResult(UPDATE_ERR_BOOTLOADER_ACCESS_RESPONSE_TIMEOUT);
					break;
				}
				updateState = UPDATE_SEND_UPDATE_CMD_TO_APP;
				mk3StartReprogram();
				vebusDeviceEnableUpdate(veTrue); /* from now on by-pass the normal frame handler */
				updateTimeout = 5 + 1;
			} else {
				veAssert(veFalse);
			}
			break;

		case UPDATE_SEND_ACCES_STRING:
			if (ev == UPDATE_TIMER)	{
				if(++retransmitCount > 100) {
					updateResult(UPDATE_ERR_BOOTLOADER_ACCESS_RESPONSE_TIMEOUT);
					return;
				}
				/* Retransmit access string */
				mk2TxBuf(bootloaderAccess, sizeof(bootloaderAccess)); /* Not a standard frame */
				updateTimeout = 2;
				return;

			} else if (ev == UPDATE_REPLY) {

				if(mk3BlVerifyRxData() == veFalse) {
					logE("mk3_upd","RX data checksum error");
					updateResult(UPDATE_ERR_PROG_RESPONSE);
					return;
				}

				if (rxBuffer[0] != BBR_VERSION)	{
					updateResult(UPDATE_ERR_VERSION_INVALID);
					logE("mk3_upd", "No version frame received");
					return;
				}

				logI("mk3_upd", "Access successful");
			} else {
				veAssert(veFalse);
				break;
			}
			/* NOTE: Fall through. We just received the version info */
		case UPDATE_GET_VERSION:
			if (ev == UPDATE_TIMER)	{
				updateResult(UPDATE_ERR_VERSION_TIMEOUT);
			} else if (ev == UPDATE_REPLY) {

				if(mk3BlVerifyRxData() == veFalse) {
					logE("mk3_upd","RX data checksum error");
					updateResult(UPDATE_ERR_PROG_RESPONSE);
					return;
				}
				logI("mk3_upd", "Got version");

				/* rxBuffer[0] = rVersionCommand ('V')*/
				/* rxBuffer[1] = cBootReprogramVersion */
				/* rxBuffer[2] = cBootFunctionsVersion */
				/* rxBuffer[3..4] = cBootBlockVersion -> (15..4 = boot block type, 3..0 = revision) */
				/* rxBuffer[5] = ExtraFlags */

				if (rxBuffer[0] != BBR_VERSION)	{
					updateResult(UPDATE_ERR_VERSION_INVALID);
					logE("mk3_upd", "No version frame received");
					return;
				}

				bootBlockVersionInfo.bootReprogramVersion = rxBuffer[1];
				bootBlockVersionInfo.bootFunctionsVersion = rxBuffer[2];
				temp = (un16) rxBuffer[3] | ((un16) rxBuffer[4] << 8);
				bootBlockVersionInfo.bootBlockType = (BootBlockType)(temp >> 4);
				bootBlockVersionInfo.bootBlockRevision = (un8) (temp & 0xF);
				bootBlockVersionInfo.extraFlags = rxBuffer[5];

				value = -1;
				vffGetParameterValueAsInt(pVffData, vffDataCPI, &value, CPT_MIN_BOOTBLOCK_FUNC);
				if (bootBlockVersionInfo.bootFunctionsVersion < value) {
					logE("mk3_upd", "Bootblock function version (%d) too old, expected version = (%d)",
						 bootBlockVersionInfo.bootFunctionsVersion, value);
					updateResult(UPDATE_ERR_VERSION_INVALID);
					return;
				}

				if (bootBlockVersionInfo.bootBlockType != MK3_BOOTBLOCK) {
					logI("mk3_upd", "Incompatible bootblock. Bootblock type = 0x%x", bootBlockVersionInfo.bootBlockType);
					updateResult(UPDATE_ERR_VERSION_INVALID);
					return;
				}

				logI("mk3_upd", "got version, change baudrate to 115200");
				retransmitCount = 0;
				updateState = UPDATE_SEND_AUTOBAUD_CMD;
				command[0] = BBC_ADAPT_BAUDRATE;
				mk3BlSendCommand(command, 1);
				updateTimeout = 10;

			} else {
				veAssert(veFalse);
			}
			break;

		case UPDATE_GET_KEY_CODE:
			if (ev == UPDATE_TIMER){
				logE("mk3_upd", "Get key code timed out");
				updateResult(UPDATE_ERR_PROG_TIMEOUT);
				break;
			} else if (ev == UPDATE_REPLY) {

				if (rxBuffer[0] != BBR_IDS)
				{
					updateResult(UPDATE_ERR_PROG_RESPONSE);
					logE("mk3_upd", "No Key codes received");
					return;
				}

				logI("mk3_upd", "got key codes");

				mk3IDs.tid = le_u32_to_cpu_p((un8 *) &rxBuffer[1]);
				mk3IDs.keyId = le_u16_to_cpu_p((un8 *) &rxBuffer[5]);
				mk3IDs.hid = le_u16_to_cpu_p((un8 *) &rxBuffer[7]);
			} else {
				veAssert(veFalse);
				break;
			}
			/* NOTE: Fall through */
		case UPDATE_SEND_UPDATE_CMD_TO_APP: /* || UPDATE_GET_KEY_CODE part 2 */
			updateState = UPDATE_SETUP_KEYS_AND_RESPONSE_TYPE;

			logI("mk3_upd", "Setup key and response type");

			setupType = 0;	/* Family key */
			if(vffGetParameterValueAsInt(pVffData, vffDataCPI, & value, CPT_USED_KEY)) {
				if (value != 0)
					setupType = 0x10; /* Private key */
			}
			/* Note: (setupType += 0x8) Tells the bootloader NOT to respond with next */
			setupType += 0x1; /* Tell the bootloader to clear program */

			wp = command;

			*wp++ = BBC_SETUP_KEY_AND_RESP_TYPE;
			*wp++ = 0;
			*wp++ = 0;
			*wp++ = (char) (mk3IDs.tid & 0xFF);
			*wp++ = (char) ((mk3IDs.tid >> 8) & 0xFF);
			*wp++ = (char) ((mk3IDs.tid >> 16) & 0xFF);
			*wp++ = (char) ((mk3IDs.tid >> 24) & 0xFF);
			*wp++ = (char) setupType;
			value = sizeof(command) - (wp - command);  /* remaining buffer size */

			value = vffGetParameterValueAsString(pVffData, vffDataCPI, strBuf, value, CPT_NONCE);

			if (value == 0) {
				logE("mk3_upd", "no Nonce found in VFF file");
				updateResult(UPDATE_ERR_VFF_FILE);
				return;
			}

			value = convertHexStringToASCII(strBuf, wp);
			wp += value;
			value = (wp - command); /* command length */

			mk3BlSendCommand(command, value);
			updateTimeout = 16;
			break;

		case UPDATE_SETUP_KEYS_AND_RESPONSE_TYPE:
			if (ev == UPDATE_TIMER){
				logE("mk3_upd", "Setup key and response type timed out");
				updateResult(UPDATE_ERR_PROG_TIMEOUT);
			} else if (ev == UPDATE_REPLY) {
				logI("mk3_upd", "Setup key and response type done");

				if(updateFromApp) {
					logI("mk3_upd", "Get version");
					updateState = UPDATE_GET_VERSION;
					command[0] = BBC_SEND_REPROGRAM_VERSION;
					mk3BlSendCommand(command,1);
					updateTimeout = 4;
				}
				else {
					logI("mk3_upd", "Get eeprom values");
					mk3BlStartGetEepromValues(command);
				}

			} else {
				veAssert(veFalse);
			}
			break;

		case UPDATE_SEND_AUTOBAUD_CMD: /* We need to change the baudrate when update is initiated from app */
			if (ev == UPDATE_TIMER){
				logI("mk3_upd", "Send autobaud string");
				updateState = UPDATE_SEND_AUTOBAUD_STRING;
				mk2SetBaudRate(MK2_BAUD_115200);
				mk2TxBuf(autoBaudString, sizeof(autoBaudString));
				updateTimeout = 4;
			} else {
				veAssert(veFalse);
			}
			break;

		case UPDATE_SEND_AUTOBAUD_STRING:
			if (ev == UPDATE_TIMER) {
				/* NOTE: autobaud does not return a version anymore.
				  This gives problems when updating fromthe VEBUS side since it it is half duplex.
				  We can get collisions, with the autobaud string and version response. */
				logI("mk3_upd", "Change baudrate done");

				if (updateFromApp) {
					logI("mk3_upd", "Get eeprom values");
					mk3BlStartGetEepromValues(command);
				}
				else {
					logI("mk3_upd", "Get keys");
					updateState = UPDATE_GET_KEY_CODE;
					command[0] = BBC_SEND_IDS;
					mk3BlSendCommand(command,1);
					updateTimeout = 4;
				}
			} else {
				veAssert(veFalse);
			}
			break;

		case UPDATE_GET_EEPROM_VALUES:
			if (ev == UPDATE_TIMER) {
				logE("mk3_upd", "Get EEPROM values timed out");
				updateResult(UPDATE_ERR_PROG_TIMEOUT);
			} else if (ev == UPDATE_REPLY) {
				if(mk3BlVerifyRxData() == veFalse || rxBuffer[0] != BBR_EEPROM_READ) {
					logE("mk3_upd","Get EEPROM values response error");
					updateResult(UPDATE_ERR_PROG_RESPONSE);
					return;
				}

				memcpy(&deviceEepromData[updateSubstateIndex*64], (char *) &rxBuffer[1], 64);

				if (++updateSubstateIndex >= 4) {
					logI("mk3_upd", "got eepromValues map, get calibration map");
					updateSubstateIndex = 0; /* Get next block of 64 bytes EEPROM data. 256 bytes total */
					updateState = UPDATE_GET_CALIBRATION_MAP;
					command[0] = BBC_SEND_CALIBRATION_LOCATIONS;
					mk3BlSendCommand(command,1);
					updateTimeout = 4;
				}
				else {
					updateState = UPDATE_GET_EEPROM_VALUES;
					command[0] = BBC_EEPROM_READ;
					command[1] = updateSubstateIndex * 64;
					command[2] = 0;
					mk3BlSendCommand(command,3);
					updateTimeout = 5;
				}
			} else {
				veAssert(veFalse);
			}
			break;

		case UPDATE_GET_CALIBRATION_MAP: /* updaterunit.pas (setupAfterConnect) */
			if (ev == UPDATE_TIMER) {
				logE("mk3_upd", "Get calibration map timed out");
				updateResult(UPDATE_ERR_PROG_TIMEOUT);
			} else if (ev == UPDATE_REPLY) {
				if(mk3BlVerifyRxData() == veFalse || rxBuffer[0] != BBR_CALIBRATION_MAP) {
					logE("mk3_upd","Get calibration map response error");
					updateResult(UPDATE_ERR_PROG_RESPONSE);
					return;
				}
				rp = mk3BlBufReadUn32(&firmwareVersionBeforeUpdate, (char*) &rxBuffer[1]);

				len = rxBufPos - 6;
				if (len > sizeof(deviceCalibrationMap)) {
					logE("mk3_upd", "Device calibration data size bigger than expected");
					updateResult(UPDATE_ERR_BUFFER_OVERFLOW);
					return;
				}

				memcpy(deviceCalibrationMap, rp, (rxBufPos - 6));

				logI("mk3_upd", "got calibration map, send EEPROM values");

				logI("mk3_upd", "firmwareVersionBeforeUpdate = %d", firmwareVersionBeforeUpdate);
				if (firmwareVersionBeforeUpdate == 0 || firmwareVersionBeforeUpdate == 0xFFFFFFFF) {
					logI("mk3_upd", "No firmware found. Complete calibration map needs to be programmed");
					if(!mk3BlMergeCalibrationData(veTrue)) { /* Clear calibration */
						updateResult(UPDATE_ERR_PROG_RESPONSE);
						 return;
					 }
				}
				else {
					if(!mk3BlMergeCalibrationData(veFalse)) {
						updateResult(UPDATE_ERR_PROG_RESPONSE);
						return;
					}
					/* NOTE: for Multi firmware update, calibration corrections are handled at this point */
				}

				/* Checking the HID and TID is performed by the bootloader. The bootloader wil return an error if not matching  */

				if(!vffGetParameterValueAsInt(pVffData, vffDataCPI, &value, CPT_EEPROM))
					value = 0;				/* No EEPROM lines in file */

				if (value != 4) {
					logE("mk3_upd","Vff file contains too many EEPROM lines (nFile = %d, nExpected = %d)", value, 4);
					updateResult(UPDATE_ERR_VFF_FILE);
					return;
				}

				updateState = UPDATE_SEND_EEPROM_VALUES;
				command[0] = BBC_EEPROM_WRITE;
				rp = pVffData->codePart[vffDataCPI].pEEPROMLines;
				rp = vffGetLineLength(&len, rp);
				memcpy(& command[1], rp, len);
				/* Insert TID. Calibration values will only be written when the correct TID is supplied */
				mk3BlBufWriteUn32 (mk3IDs.tid, & command[3]);

				pVffData->pRead = rp + len;

				updateSubstateIndex = 0;
				mk3BlSendCommand(command, len + 1); /* Send first line */
				updateTimeout = 16;	/* EEPROM write cycle time -> 4 ms * 64 = 256ms + overhead, erase */
			} else {
				veAssert(veFalse);
			}
			break;

		case UPDATE_SEND_EEPROM_VALUES:
			if (ev == UPDATE_TIMER)
			{
				logE("mk3_upd", "EEPROM programming timed out");
				updateResult(UPDATE_ERR_PROG_TIMEOUT);
			} else if (ev == UPDATE_REPLY) {
				if((mk3BlVerifyRxData() == veFalse) || (rxBuffer[0] != BBR_NEXT)) {
					logE("mk3_upd","EEPROM write response error");
					updateResult(UPDATE_ERR_PROG_RESPONSE);
					return;
				}

				if (++updateSubstateIndex >= 4) {
					logI("mk3_upd", "EEPROM programming done, send program code");

					if(!vffGetParameterValueAsInt(pVffData, vffDataCPI, &value, CPT_CODE))
						value = 0;				/* No code lines in file */

					updateSubstateIndex = value; /* Count back */

					rp = pVffData->codePart[vffDataCPI].pCodeLines;
					rp = vffGetLineLength(&len, rp);

					memcpy(& command[1], rp, len);
					pVffData->pRead = rp + len;

					updateState = UPDATE_SEND_CODE;

					command[0] = BBC_PROGRAM_WRITE;
					mk3BlSendCommand(command,len + 1);
					updateTimeout = 32;	/* Flash write cycle time -> 2 ms * 64 = 128ms + overhead */
				}
				else {
					updateState = UPDATE_SEND_EEPROM_VALUES;
					command[0] = BBC_EEPROM_WRITE;
					rp = vffGetLineLength(&len, pVffData->pRead);
					pVffData->pRead = rp + len;
					memcpy(& command[1], rp, len);
					mk3BlBufWriteUn32 (mk3IDs.tid, & command[3]);
					mk3BlSendCommand(command, len + 1); /* Send next */
					updateTimeout = 16;	/* EEPROM write cycle time -> 4 ms * 64 = 256ms + overhead, erase */
				}
			} else {
				veAssert(veFalse);
			}
			break;

		case UPDATE_SEND_CODE:
			if (ev == UPDATE_TIMER)
			{
				logE("mk3_upd", "Code programming timed out");
				updateResult(UPDATE_ERR_PROG_TIMEOUT);
			} else if (ev == UPDATE_REPLY) {
				if((mk3BlVerifyRxData() == veFalse) || (rxBuffer[0] != BBR_NEXT)) {
					logE("mk3_upd","Code write response error");
					updateResult(UPDATE_ERR_PROG_RESPONSE);
					return;
				}

				if (--updateSubstateIndex <= 0) {
					logI("mk3_upd", "Code programming done, start MK3 application");

					updateState = UPDATE_RUN_APPLICATION;

					command[0] = BBC_RUN_RESET;
					mk3BlSendCommand(command, 1);
					updateTimeout = 16;
				}
				else {
					updateState = UPDATE_SEND_CODE;
					command[0] = BBC_PROGRAM_WRITE;
					rp = vffGetLineLength(&len, pVffData->pRead);
					pVffData->pRead = rp + len;
					memcpy(& command[1], rp, len);

					mk3BlSendCommand(command,len + 1);
					updateTimeout = 32;	/* Flash write cycle time -> 2 ms * 64 = 128ms + overhead */
				}
			} else {
				veAssert(veFalse);
			}
			break;

		case UPDATE_RUN_APPLICATION:
			if (ev == UPDATE_TIMER)
			{
				/* If the program checksum fails the bootloader will send a reprogram version frame */
				logI("mk3_upd", "Program start done");
				/* Done */
				updateResult(0);
				return;
			} else if (ev == UPDATE_REPLY) {
				if((mk3BlVerifyRxData() == veFalse)) {
					logE("mk3_upd","Run program response error");
					updateResult(UPDATE_ERR_PROG_RESPONSE);
					return;
				}

				logE("mk3_upd", "Program start failed");
				updateResult(UPDATE_ERR_PROG_RESPONSE);
				/* Done */
				return;
			} else {
				veAssert(veFalse);
			}
			break;
	}
}

#ifndef CFG_MK2_BAUDRATE		/* The MK3 can run at higher baudrates */
#define CFG_MK2_BAUDRATE	2400
#endif
/* all exit paths should go through here */
static void updateResult(UpdateError error)
{
	updateError = error;

	updateTimeout = 0;	/* no more timer events */
	bytesExpected = 0;	/* no data events.. */

	logI("mk3_upd", (error == UPDATE_ERR_OK ? "ok" :  "error" ));

	if ((error == UPDATE_ERR_BOOTLOADER_ACCESS_RESPONSE_TIMEOUT) && (++ updateRetryCount <= 2)) {

		updateState = UPDATE_START;
		if (updateRetryCount == 1){
			logI("mk3_upd", "Update from application failed, retry from bootloader");
			mk3FirmwareUpdateState(UPDATE_BEGIN_FROM_BOOTLOADER);
			return;
		}
		else if (updateRetryCount == 2) {
			logI("mk3_upd", "Update from application failed, retry from bootloader");
			mk3FirmwareUpdateState(UPDATE_BEGIN_WITH_RESET);
			return;
		}
	}

	updateState = UPDATE_IDLE;

	/* restore normal mk3 serial settings */
	mk2SetBaudRate(CFG_MK2_BAUDRATE);
	vebusDeviceEnableUpdate(veFalse);

	if (resultCallback) {
		resultCallback(error == UPDATE_ERR_OK ? VBDUP_OK : VBDUP_UPDATE_FAILED );
		resultCallback = NULL;
	}
}

/**
 * Start updating, eventually callback will be invoked with true/false
 *
 * Start update from application or from bootloader.
 * When there is no (compatible) application the START_UPDATE_FROM_BOOTLOADER mode should be selected.
 * In this mode the MK3 should be reset.
 *
 * @param mode Start update from bootloader or application
 * @param pData VFF data buffer
 * @param dataIndex VFF data buffer codepart index
 * @param callback callback when done
 */
void mk3FirmwareUpdate(Mk3UpdateMode mode, VffData* pData, int dataIndex, VbdubUpdateResultCallback* callback)
{
	veAssert(updateState == UPDATE_IDLE);
	updateState = UPDATE_START;
	resultCallback = callback;
	pVffData = pData;
	vffDataCPI = dataIndex;

	updateRetryCount = 0;

	switch (mode)
	{
		case START_UPDATE_FROM_BOOTLOADER:
			mk3FirmwareUpdateState(UPDATE_BEGIN_FROM_BOOTLOADER);
			break;
		case START_UPDATE_FROM_APPLICATION:
			mk3FirmwareUpdateState(UPDATE_BEGIN_FROM_APP);
			break;
		default:
		if (resultCallback) {
		resultCallback(VBDUP_UPDATE_FAILED);
		resultCallback = NULL;
		}
	}
}

static void getIDsDone(void)
{
	logI("mk3_upd", "getIDsDone");
	mk3FirmwareUpdateState(UPDATE_REPLY);
}

/* 50ms, main */
void mk3FirmwareTick(void)
{
	if (updateTimeout && --updateTimeout == 0)
		mk3FirmwareUpdateState(UPDATE_TIMER);

	if (dataReady)
	{
		dataReady = veFalse;
		mk3FirmwareUpdateState(UPDATE_REPLY);
	}
}


/**
 * on interrupt! Called from mk2_serial.c
 *
 * @param inbyte Received byte
 */
void mk3RxUpdateRxByte(un8 inbyte)
{
	if(!dataExpected)
		return;					/* ignore incommming data if we don't expect data */

	if (receivedStartChar) {
		logI("mk3_upd", "receivedStartChar");
		if (inbyte & 0x80) { /* bit 7 must be set */
			receivedStartChar = veFalse;
			bytesExpected = (inbyte & 0x7F) + 1; /* +1 for checksum */
			logI("mk3_upd", "bytesExpected = %d", bytesExpected);
		}
		return;
	}
	else if (bytesExpected == 0 && inbyte == 0) {
		receivedStartChar = veTrue;
		return;
	}

	if (bytesExpected == 0)
		return;

	/* Receiving a bootblock frame */

	if(receivedEscapeChar && inbyte == 0x80)
		rxBuffer[rxBufPos++] = 0;
	else if (inbyte == 0) {
		/* The next byte should be 0x8 */
		receivedEscapeChar = veTrue;
		return;
	}
	else
		rxBuffer[rxBufPos++] = inbyte;

	/* logI("mk3_upd","rxBufPos = %d, data = 0x%x", rxBufPos, (un8) rxBuffer[rxBufPos-1]); */

	receivedEscapeChar = veFalse;

	if (rxBufPos >= sizeof(rxBuffer)) {
		logE("mk3_upd", "input rxBuffer overflow");
		bytesExpected = 0;
		rxBufPos = 0;
		return;
	}

	/* If data was expected, let the main loop pick it up.. */
	if (--bytesExpected == 0)
	{
		logI("mk3_upd","dataReady");

		dataReady = veTrue;
		dataExpected = veFalse;
	}
}

/* Returns if update is busy, to suspend normal behaviour */
veBool mk3UpdateBusy(void)
{
	return (updateState != UPDATE_IDLE);
}


#if 0							/* Not used, keep for reference */
/**
 * Extract the HID from the key code
 *
 * @param pKeyCode
 *
 * @return HID
 */
static un16 mk3BlHIDFromKey(char* pKeyCode)
{
	return ((pKeyCode[6] << 8) || pKeyCode[7]);
}

/**
 * Extract the TID from the key code
 *
 * @param pKeyCode
 *
 * @return TID
 */
static un32 mk3BlTIDFromKey(char* pKeyCode)
{
	un32 tid;

	mk3BlBufReadUn32(&tid,pKeyCode);

	return tid;
}
#endif
