#ifndef _VELIB_INC_MK2_H_
#define _VELIB_INC_MK2_H_

#include <velib/velib_config_app.h>
#include <velib/mk2/frame_handler.h>
#include <velib/mk2/mk2.h>
#include <velib/mk2/mk2_actions.h>
#include <velib/mk2/values_float.h>
#include <velib/mk2/vebus_device.h>
#include <velib/mk2/vebus_system.h>
#include <velib/mk2/vsc.h>
#include <velib/platform/serial.h>

/* Tweaked version of some structs to minimize memory usage a bit for the CAN cable */

typedef enum {
	UNKNOWN = 0,
	MK2 = 2,
	MK3 = 3
} Mk2Type;

typedef struct Mk2S
{
	/* version of the mk2 itself */
	un32 version;
	/* state of the mk2 itself */
	Mk2State state;
	/* variables for the mk2 frame handler */
	Mk2FhVars fh;
	/* last address the mk2 is going to set to */
	un8 intendedAddress;
	/* device the mk2 is set to talk to */
	un8 actualAddress;
	/* last command send to support XYZ looping */
	un8 cmd;
	/* mk2 product model */
	Mk2Type type;
} Mk2;

/*
 * Since there is not enough memory to keep the scales per device,
 * only the scales from the master is used, assuming it is equal
 * to the slaves.
 */
typedef struct {
	/* Scale information of this device */
	VebusSettingInfo settingInfo[VEBUS_SETTINGS_COUNT];
	VebusRamInfo ramInfo[VEBUS_RAMVAR_COUNT];
	veBool validScales;
} VebusScales;

typedef struct {
	veBool setExtraBatteryCurrent;
	sn32 extraBatteryCurrent;
} SocControl;

/*
 * Readings of the device and their timeouts
 * VEBUS_RAMVAR_COUNT is enough for reading known ramvars alone.
 * When reading assistant variables, they are placed in the same array
 * but there is big hole between them, since assistant ids start at 0x80.
 */
#ifdef CFG_MK2_ASSISTANT_VARS_COUNT
#define VEBUS_RAMASS_VARS_LAST	(0x80 + CFG_MK2_ASSISTANT_VARS_COUNT)
#else
#define VEBUS_RAMASS_VARS_LAST	VEBUS_RAMVAR_COUNT
#endif

typedef struct VebusDeviceS
{
	/* verion of the vebus device */
	VebusVersion version;
	un32 uniqueNumber;
	un8 type;
	/* vebus short address used by this device */
	un8 addr;
	/* Readings of the device */
	un16 vars[VEBUS_RAMASS_VARS_LAST];
	/* Settings of the device, these won't timeout */
	un16 settings[VEBUS_SETTINGS_COUNT];
	un8 settingValidFlags[MK2_ACT_SETTINGS_FLAG_BYTES];
	/* Queued actions */
	Mk2DevActVars act;
	/* vebus access level support */
	veBool resetRequired;
	un8 accessLevelRequired;

	/* Linked list */
	struct VebusDeviceS* next;
} VebusDevice;

typedef struct VebusSystemS
{
	/* not used */
	VeUnitSet units;
	/* State of the whole system */
	VebusState state;
	/* Workaround / features which might need to be considered */
	un16 quirks;
	/* derived alarms from state.alarms and the leds */
	un16 alarms;
	/* Charge and / or inverter state */
	VebusConverterState convState;
	/* DC values raw  */
	Mk2RawDc dc;
	/* Reported current limit as (last) used by the whole system for the active input. */
	VebusAcInConfig acInConfig;
	/* Copy of the above to keep the min/max available if it is not the active input */
	VebusAcInConfig acInConfigPerAcIn[2];
	/*
	 * AC values for three phases. The first phase also contains
	 * the number of inverters / chargers involved.
	 */
	Mk2RawAcPhase ac[3];
	/* Variables needed for quering scales of a VEbus device */
	VebusInitVars init;
	/* pending actions */
	Mk2SystemActVars act;
	/* One mk2 to interface a VEBus system */
	struct Mk2S mk2Interface;
	/* Port the mk2 is connected to */
	VeSerialPort port;
	/* One vebus system can consist out of multi devices */
	struct VebusDeviceS *devices;
	/* The master device (if found) */
	struct VebusDeviceS *master;
	/* kwh related variables */
	VeKwhVars kwh;

	Mk2ActIdle idle;
	Mk2ActIdle* idleCb;

	/* Variables needed for controlling on / off and current limit */
	VebusControl ctrl;

	VebusPower power;

	/*
	 * The vebus system can optionally contain assistants sending
	 * out the voltage setpoint. This is forwardingen to the CAN-bus in
	 * for other chargers to follow.
	 */
	un16 voltageSetpoint;

	VeCurrentSensorVars iSense;
	VeHub1Vars hub1;
	VeAssistantVars assistant;

	veBool validRuntimeValues;

	/* AUX devices verions (Digital Multi Control and Battery Management System) */
	un32 bmsVersion;
	un32 dmcVersion;

	SocControl socControl;

	Mk2Actions nextAction;
} VebusSystem;

extern struct VebusSystemS		vebusSystem;
extern struct Mk2S				mk2;

/* There is not enough memory to load all scales so use only one copy */
extern VebusScales		vebusScales;

#endif
