#include <string.h> // memcpy
#include <stdio.h>

#include <velib/velib_config.h>

#ifdef NMEA2K
#include <velib/velib_inc_J1939.h>
#include <velib/velib_inc_vecan.h>
#endif

#include <velib/J1939/J1939_device.h>
#include <velib/nmea2k/n2k.h>
#include <velib/mk2/vebus_system.h>
#include <velib/utils/ve_logger.h>
#include <velib/vecan/prod_data.h>
#include <velib/vecan/prod_data_cm.h>

#include "../boot/src/product.h"
#include <dev_reg.h>
#include <platform.h>
#include <storage.h>

DevRegStored devRegStored;
static veBool writeSettings;

#define REG_FORMAT			5

DevRegStored const devRegStoredDefault =
{
	0x10,					// NAD F0 (output).

	REG_FORMAT,				// format;
	sizeof(DevRegStored),	// length;
	0,						// checksum

	0,						// Device flags
	SP_SWITCH_ON,
	{						// Current Limits
		0xFFFF,
		0xFFFF,
		0xFFFF,
		0xFFFF,
	},

	0,						// System Instance F0.
	0,						// Device Instance F0.

	0x11,					// NAD F1 (input).
	1,						// Device Instance F1.

	0,						// Battery Instance.
	0,						// Type -> battery.
	1,						// Starter Instance.

	{"Multi - Inverter / Charger"},		// Config field 1.
	{""},								// Config field 2.

	{"Multi - AC Input"},				// Config field 1.
	{""},								// Config field 2.

	// Message settings
	{
#ifdef NMEA2K
		{0,		6},		// MSG_F0_ACL
		{0,		6},		// MSG_F0_CONFIG_INFO
		{0,		6},		// MSG_F0_PROD_INFO
		{0,		6},		// MSG_F0_PNGS_TX
		{0,		6},		// MSG_F0_PNGS_RX

		{30,	6},		// MSG_F0_BATTERY_BANK
		{0,		6},		// MSG_F0_DC_DETAILED
		{30,	6},		// MSG_F0_INVERTER_STATE
		{30,	6},		// MSG_F0_CHARGER_STATE
		{30,	6},		// MSG_F0_SWITCH_BANK
		{100,	6},		// MSG_F0_SWITCH_BANK_LEDS

		{30,	6},		// MSG_F0_AC_OUT1_P1
		{30,	6},		// MSG_F0_AC_OUT1_P1_POWER
		{30,	6},		// MSG_F0_AC_OUT1_P2
		{30,	6},		// MSG_F0_AC_OUT1_P2_POWER
		{30,	6},		// MSG_F0_AC_OUT1_P3
		{30,	6},		// MSG_F0_AC_OUT1_P3_POWER
		{30,	6},		// MSG_F0_AC_OUT1_TOTAL_POWER

#ifdef CFG_N2K_PROPOSAL
		{0,		6},		// MSG_F0_N2K_AC_IN_V1
		{0,		6},		// MSG_F0_N2K_AC_IN_P1
		{0,		6},		// MSG_F0_N2K_AC_IN_V2
		{0,		6},		// MSG_F0_N2K_AC_IN_P2
		{0,		6},		// MSG_F0_N2K_AC_IN_V3
		{0,		6},		// MSG_F0_N2K_AC_IN_P3

		{0,		6},		// MSG_F0_N2K_AC_OUT_V1
		{0,		6},		// MSG_F0_N2K_AC_OUT_P1
		{0,		6},		// MSG_F0_N2K_AC_OUT_V2
		{0,		6},		// MSG_F0_N2K_AC_OUT_P2
		{0,		6},		// MSG_F0_N2K_AC_OUT_V3
		{0,		6},		// MSG_F0_N2K_AC_OUT_P3

		{0,		6},		// MSG_F0_STATE,
#endif
		{0,		6},		// MSG_F1_ACL
		{0,		6},		// MSG_F1_CONFIG_INFO
		{0,		6},		// MSG_F1_PROD_INFO
		{0,		6},		// MSG_F1_PNGS_TX
		{0,		6},		// MSG_F1_PNGS_RX

		{30,	6},		// MSG_F0_AC_IN1_P1
		{30,	6},		// MSG_F0_AC_IN1_P1_POWER
		{30,	6},		// MSG_F0_AC_IN1_P2
		{30,	6},		// MSG_F0_AC_IN1_P2_POWER
		{30,	6},		// MSG_F0_AC_IN1_P3
		{30,	6},		// MSG_F0_AC_IN1_P3_POWER
		{30,	6},		// MSG_F0_AC_IN1_TOTAL_POWER
#else
		{0,		0},		// Dummy
#endif
	},

	0,	/* alarmEventInstance */
	1,	/* ledEventInstance */
	0,	/* BMS Version */
	0,	/* DMC Version */
};


void devRegInit(void)
{
	veBool valid;

	writeSettings = veFalse;

	valid = storageLoad();

	if (valid && devRegStored.length == 0)
		valid = veFalse;


	if (valid) {
		/* adjust format to new newer version */
		if (devRegStored.format == 4) {
			devRegStored.bmsVersion = 0;
			devRegStored.dmcVersion = 0;
			devRegStored.format = 5;
			storageSave();
		}

		/* force defaults if incompatible version.. */
		if (devRegStored.format != REG_FORMAT) {
			logI("", "invalid format");
			valid = veFalse;
		}
	}

	// set default if not valid.
	if (!valid)
	{
		logI("", "restoring default");
		devRegDefault();
		storageSave();
	}

	devRegApply();
}

// Copy settings to the correct place.
void devRegApply(void)
{
#ifdef NMEA2K
	// Create ACL name. Part is product specific and fixed (serial, type).
	// Fixed values are stored in the bootloader...
	j1939Devices[0].nad = devRegStored.nad;
	j1939Devices[0].name.low = prodInfo.nameLow;
	j1939Devices[0].name.high.vun32 = prodInfo.nameHigh;

	// Instance can be programmed.
	j1939Devices[0].name.high.vun32 &= ~0xFF;
	j1939Devices[0].name.high.fields.sca = 1;
	j1939Devices[0].name.high.vun32 |= devRegStored.deviceInstanceF0;
	j1939Devices[0].name.high.fields.dcInst = devRegStored.systemInstance;

	// Additional function...
	j1939Devices[1].nad = devRegStored.nadF1;
	j1939Devices[1].name.low = prodInfo.nameLow+1;

	// Second function reporst AC Input values..
	j1939Devices[1].name.high.vun32 =
		J1939_ACL_H_SCA | J1939_ACL_H_INDUSTRY(J1939_INDUSTRY_MARINE) |
		J1939_ACL_H_CLASS_INST(devRegStored.systemInstance) |
		J1939_ACL_H_CLASS(N2K_CLASS_ELECTRICAL_GENERATION) |
		J1939_ACL_H_FUNCTION(N2K_FUNCTION_AC_INPUT) | devRegStored.deviceInstanceF1;
#endif
}

void devRegDefault(void)
{
	memcpy(&devRegStored, &devRegStoredDefault, sizeof(devRegStoredDefault));
}
