#include <velib/utils/ve_logger.h>

#include "storage.h"
#include "dev_reg.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

static char const *settingsFile = "settings.bin";

static un16 getCheckSum(void)
{
	un8 const* ptr;
	un16 checksum;

	// just sum all bytes in the settings..
	checksum = 0;
	ptr = (un8 const*) &devRegStored + sizeof(devRegStored);
	while (ptr > (un8 const*) &devRegStored)
		checksum += *--ptr;

	// undo adding the current checksum..
	checksum -= (un8) devRegStored.checksum;
	checksum -= (un8) (devRegStored.checksum >> 8);

	// do not allow 0xFFFF as valid checksum.
	if (checksum == 0xFFFF)
		checksum--;

	return checksum;
}

/**
 * Load the settings from a file..
 *
 * @returns wether the settings could be succesfully loaded and are valid.
 */
veBool storageLoad(void)
{
	FILE* fh;
	un16 length;

	fh = fopen(settingsFile, "rb");
	if (fh == NULL)
		return veFalse;

	if (fseek(fh, offsetof(DevRegStored, length), SEEK_SET) != 0) {
		logE("storage", "fseek failed");
		fclose(fh);
		return veFalse;
	}

	if (fread((un8*) &length, 1, sizeof(length), fh) != sizeof(length)) {
		logE("storage", "fread length failed");
		fclose(fh);
		return veFalse;
	}

	if (length > sizeof(devRegStored)) {
		logE("storage", "invalid length");
		fclose(fh);
		return veFalse;
	}

	if (fseek(fh, 0, SEEK_SET) != 0) {
		logE("storage", "fseek begin failed");
		fclose(fh);
		return veFalse;
	}

	if (fread((un8*)&devRegStored, 1, length, fh) != length) {
		logE("storage", "fread data failed");
		fclose(fh);
		return veFalse;
	}

	fclose(fh);

	if (getCheckSum() != devRegStored.checksum) {
		logE("storage", "invalid checksum");
		return veFalse;
	}

	return veTrue;
}

/// Store settings to a file.
void storageSave(void)
{
	int ret,err;
	FILE* fh;
	char tempFileName[255];

	devRegStored.checksum = getCheckSum();

	/* Use the same dir for the tmp file and settingsFile.
	 * Rename can't act on different partitions. */
	strncpy(tempFileName, settingsFile, sizeof(tempFileName));
	strncat(tempFileName, ".tmp", sizeof(tempFileName));

	fh = fopen(tempFileName, "wb");

	if (fh == NULL)
		return;
	fwrite((un8*)&devRegStored, 1, sizeof(DevRegStored), fh);
	fclose(fh);

	sync();
	ret = rename(tempFileName, settingsFile);
	if (ret != 0) {
		err = errno;
		logE("storage", "Rename %s failed with errno = %d", tempFileName, err);
	}

	sync();
}

void storageTick(void)
{
}


void storageSetFilename(const char *path)
{
	settingsFile = path;
}
