#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <velib/velib_inc_mk2.h>

#include <velib/base/base.h>
#include <velib/base/ve_string.h>
#include <velib/mk2/hal.h>
#include <velib/mk2/mk2_actions.h>
#include <velib/mk2/mk2_update.h>
#include <velib/mk2/vebus_device.h>
#include <velib/mk2/vebus_system.h>
#include <velib/mk2/vebus_aux_device.h>
#include <velib/vecan/regs.h>
#include <velib/vecan/products.h>
#include <velib/utils/ve_arg.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_logger.h>
#include <velib/utils/ve_timer.h>
#include <velib/utils/ve_todo.h>
#include <velib/platform/console.h>
#include <velib/platform/plt.h>
#include <velib/vebusfr/vebusfr.h>
#include <platform.h>
#include <bus_values.h>
#include <dev_reg.h>
#include <msg_router.h>
#include <product_info.h>
#include <storage.h>
#include <task.h>

#if CFG_WITH_MK3 == 1
#include <velib/mk3/mk3.h>
#include <mk3_firmware_upgrade.h>
#endif


typedef enum {
	PANEL_IDLE,
	PANEL_QUEUED,
	PANEL_BUSY,
} PanelState;

static un8 startupTimer;
static un16 kwhTimer;
static veBool getKwhValues = veTrue;
static veBool allowMk2Update;
static veBool initDone;
static PanelState sendPanelFrame;
static PanelState sendSoc;
static un16 socToWrite;
static mk2TakeOverRequestCb *mk2TakeOverCb;
static veBool mk2IsTakenOver;
static Mk2Callback *delayCb;
static un8 delayCount;
static Mk2Type mk2TypeInit = MK2;
/*
 * The actions RAM var is bypassed since it might wait for all
 * to be requested, but only the SOC of charge it interesting to request
 * periodically.
 */
#define MK2_ACT_WRITE_SOC 97
#define VEBUS_GET_POWER	98
#define MK2_ACT_VOLTAGE_SETPOINT 99
#define MK2_ACT_SOC 100

static void getSocDone(void);
static void getVoltageSetpointDone(void);
static void qeuryDone(void);
static void mk2UpdateDone(Mk2Result result);
static void mk2Begin(void);

static void queueNextAction(VebusDevice* dev, Mk2Actions done);

#if 0
static veBool mk2TakenOver(void)
{
	return mk2TakeOverCb != NULL || mk2IsTakenOver;
}
#endif

static void getDcDone(void)
{
	queueNextAction(NULL, MK2_ACT_DC_VALUES);
}

static void getVebusStateDone(void)
{
	queueNextAction(NULL, MK2_ACT_STATUS);
}

static void getConvStateDone(void)
{
	queueNextAction(NULL, MK2_ACT_CONV_STATE);
}

static void getAssistantValuesDone(void)
{
	queueNextAction(NULL, MK2_ACT_GET_ASSISTANT_VALUES);
}

static void reqAcInConfigDone(void)
{
	queueNextAction(NULL, MK2_ACT_AC_CONFIG);
}

static void getKwhFromDeviceDone(void)
{
	queueNextAction(NULL, MK2_ACT_GET_KWH_COUNTERS);
}

#if CFG_WITH_CONTROL
static void sendPanelFrameDone(void)
{
	queueNextAction(NULL, MK2_ACT_SET_PANEL_STATE);
}
#endif

static void sendIBatExtraDone(void)
{
	queueNextAction(NULL, MK2_ACT_SET_IBAT_EXTRA_OR_SOC);
}


/* Get next phase if any or start over */
static void getAcDone(void)
{
	if ((vebusSystem.act.varId + 1) >= vebusSystem.ac[0].phaseCount)
	{
		vebusSystem.act.varId = 0;
		queueNextAction(NULL, MK2_ACT_AC_VALUES);
		return;
	}

	vebusSystem.act.varId++;
	vebusGetAcPhase(vebusSystem.act.varId, &vebusSystem.ac[vebusSystem.act.varId], getAcDone);
}

static void getVoltageSetpointDone(void)
{
	vebusSystem.voltageSetpoint = vebusSystem.act.varValue;
	queueNextAction(vebusSystem.master, MK2_ACT_VOLTAGE_SETPOINT);
}

static void getSocDone(void)
{
	logI("mk2", "got SOC");
	vebusSystem.master->vars[VE_VAR_SOC_RAMVAR] = vebusSystem.act.varValue;
	valuesRamVarObtainedCallback(vebusSystem.master, VE_VAR_SOC_RAMVAR);
	queueNextAction(NULL, MK2_ACT_SOC);
}

static veBool powerControlAllowed(void)
{
	if (!vebusCanBeControlled() || devRegStored.mode == SP_SWITCH_OFF)
		return veFalse;

	return veTrue;
}

static void invalidate(void)
{
	VebusDevice* next;
	VebusState preservedState;

	delayCount = 0;

	while (vebusSystem.devices) {
		next = vebusSystem.devices->next;
		vebusDeviceFree(vebusSystem.devices);
		vebusSystem.devices = next;
	}
	vebusSystem.master = NULL;
	vebusSystem.state.knownIds = 0;
	preservedState = vebusSystem.state;
	memset(&vebusSystem, 0, sizeof(VebusSystem));
	vebusSystem.state = preservedState;
	vebusSystem.bmsVersion = devRegStored.bmsVersion;
	vebusSystem.dmcVersion = devRegStored.dmcVersion;
	valuesInvalidate();

	mk2Init();
	mk2ActInit();
	mk2FhReset();
	vebusControlUpdate();

	/* CAN bus specific values */
	vebusSystem.voltageSetpoint = 0xFFFF;

	mk2IsTakenOver = veFalse;
	sendPanelFrame = PANEL_IDLE;
	initDone = veFalse;
}

static void powerSnapShotDone(void)
{
	queueNextAction(vebusSystem.master, VEBUS_GET_POWER);
}

static void writeSocDone(void)
{
	queueNextAction(NULL, MK2_ACT_WRITE_SOC);
}

/**
 * Check if there is some other action queued for any device
 * returns true when doing something else
 * return false when completly idle
 */
static veBool sendOtherMessage(void)
{
	veBool idle = vebusSystem.act.dev == 0 || mk2ActUpdate(vebusSystem.act.dev);

	if (idle) {
		logI("task", "device was done, checking other devices");
		VebusDevice *dev = vebusSystem.devices;
		while (dev) {
			if (mk2ActUpdate(dev) == veFalse)
				return veTrue;
			dev = dev->next;
		}
	} else {
		return veTrue;
	}

	logI("task", "nothing other to do");

	return veFalse;
}

/*
 * Periodic actions when started.
 *
 * There are a couple of different messages needed to be send. First the mk2 / vebus
 * devices need to be found and scaling needs to be obtained. This is done in
 * vebus_device_init. During this time no runtime data is requested (indicated by
 * vebusScales.validScales). Panel frames are sent though, but that is done by the
 * vebusSystem code itself. Thereafter runtime information is requested one by one.
 * If there are other messages queued, like changing setting etc, these are allowed to
 * go in between, but only one at the time, to also keep the runtime data updated.
 *
 * The intention is to create something like below, the panel frames (p) are send
 * periodically. In between is the looping requests, a b c. Each of these request might
 * be followed by another messages (s).
 *
 *   p					p				p
 *		a	s?	b	s?		c	s?	a		s?	b
 */
void queueNextAction(VebusDevice* dev, Mk2Actions done)
{
	static veBool allowOtherMessage = veTrue;
	VE_UNUSED(dev);

	if (sendSoc == PANEL_BUSY) {
		veAssert(done == MK2_ACT_WRITE_SOC);
		sendSoc = PANEL_IDLE;

	/* Panel frames can allows come in between periodic messages */
	} else if (sendPanelFrame == PANEL_BUSY) {
		veAssert(done == MK2_ACT_SET_PANEL_STATE);
		sendPanelFrame = PANEL_IDLE;

	/*
	 * Optional messages like settings etc might go in between the
	 * known messages. Since only 1 message is allowed, allowOtherMessage
	 * is false when an optional message has been send.
	 */
	} else if (allowOtherMessage == veFalse) {
		logI("mk2", "other message is done %d\n=========================================================", done);

	/* Periodic messages, the response to these messages are well known */
	} else {
		veAssert(vebusSystem.nextAction == done);
		logI("mk2", "periodic msg is done %d\n=========================================================", done);

		switch ((int) done) {
		case MK2_ACT_AC_CONFIG:
			vebusSystem.nextAction = MK2_ACT_STATUS;
			break;
		case MK2_ACT_STATUS:
			vebusSystem.nextAction = MK2_ACT_CONV_STATE;
			break;
		case MK2_ACT_CONV_STATE:
			vebusSystem.nextAction = MK2_ACT_DC_VALUES;
			break;
		case MK2_ACT_DC_VALUES:
			vebusSystem.nextAction = VEBUS_GET_POWER;
			break;
		case VEBUS_GET_POWER:
			vebusSystem.nextAction = MK2_ACT_AC_VALUES;
			break;
		case MK2_ACT_AC_VALUES:
			vebusSystem.nextAction = (getKwhValues ? MK2_ACT_GET_KWH_COUNTERS : MK2_ACT_GET_ASSISTANT_VALUES);
			break;
		case MK2_ACT_GET_KWH_COUNTERS:
			/*
			 * If there are more device to query this state is not yet left.
			 * This is done here and not at a lower level, to allow other message
			 * to go in between.
			 */
			if (!vebusKwhMoreTodo()) {
				getKwhValues = veFalse;
				kwhTimer = 30 * 20;
				vebusSystem.nextAction = MK2_ACT_GET_ASSISTANT_VALUES;
			}
			break;
		case MK2_ACT_GET_ASSISTANT_VALUES:
			/*
			 * There might be multiple sensors. Update them all..
			 */
			if (!vebusCurrentSensorsMoreTodo()) {
				if (vebusSystem.hub1.ramVarOffset != 0)
					vebusSystem.nextAction = MK2_ACT_VOLTAGE_SETPOINT;
				else
					vebusSystem.nextAction = MK2_ACT_SOC;
			}
			break;
		case MK2_ACT_VOLTAGE_SETPOINT:
			vebusSystem.nextAction = MK2_ACT_SOC;
			break;
		case MK2_ACT_SOC:
			vebusSystem.nextAction = MK2_ACT_SET_IBAT_EXTRA_OR_SOC;
			break;
		case MK2_ACT_SET_IBAT_EXTRA_OR_SOC:
			vebusSystem.validRuntimeValues = veTrue;
			vebusSystem.nextAction = MK2_ACT_AC_CONFIG;
			break;
		default:
			veAssert(veFalse);
		}

		valuesActionDoneCallback(done);

		logI("mk2", "next action %d", vebusSystem.nextAction);
	}

	if (sendSoc == PANEL_QUEUED) {
		if (vebusSupportedIBatSocBroadcast(vebusSystem.master)) {
			logI("mk2", "broadcast SOC raw %d", socToWrite);
			vebusFrSendSoc(writeSocDone, socToWrite);
		}
		else {
			logI("mk2", "sending SOC raw %d", socToWrite);
			vebusSetVar(vebusSystem.master, VE_VAR_SOC_RAMVAR, writeSocDone, socToWrite);
		}

		sendSoc = PANEL_BUSY;
		return;
	}

#if CFG_WITH_CONTROL
	/* Panel frames can always be send in between normal traffic */
	if (sendPanelFrame == PANEL_QUEUED) {
		logI("mk2", "sending panel frame");
		sendPanelFrame = PANEL_BUSY;
		vebusSendPanelFrame(sendPanelFrameDone);
		return;
	}
#endif

	/* Schedule optional messages in between, but only once */
	if (allowOtherMessage && sendOtherMessage()) {
		allowOtherMessage = veFalse;
		logI("task", "allow other message");

	/* periodic requests */
	} else {
		logI("mk2", "do next action %d", vebusSystem.nextAction);
		allowOtherMessage = veTrue;

		switch ((int) vebusSystem.nextAction) {
		/* provide breakpoints */
		case MK2_ACT_AC_CONFIG:
			vebusGetAcInConfig(&vebusSystem.acInConfig, reqAcInConfigDone);
			break;
		case MK2_ACT_STATUS:
			mk2GetVebusState(&vebusSystem.state, getVebusStateDone);
			break;
		case MK2_ACT_CONV_STATE:
			vebusGetConverterState(vebusSystem.master, getConvStateDone, &vebusSystem.convState);
			break;
		case MK2_ACT_DC_VALUES:
			vebusGetDcValues(&vebusSystem.dc, getDcDone);
			break;
		case VEBUS_GET_POWER:
			if (vebusGetPower(powerSnapShotDone))
				break;
			vebusSystem.nextAction = MK2_ACT_AC_VALUES;
			// fall through
		case MK2_ACT_AC_VALUES:
			vebusSystem.act.varId = 0; // phase
			vebusGetAcPhase(vebusSystem.act.varId, &vebusSystem.ac[0], getAcDone);
			break;
		case MK2_ACT_GET_KWH_COUNTERS:
			if (vebusKwhUpdate(getKwhFromDeviceDone))
				break;
			vebusSystem.nextAction = MK2_ACT_GET_ASSISTANT_VALUES;
			// fall through
		case MK2_ACT_GET_ASSISTANT_VALUES:
			if (vebusGetCurrentSensors(getAssistantValuesDone))
				break;
			vebusSystem.nextAction = MK2_ACT_VOLTAGE_SETPOINT;
			// fall through
		case MK2_ACT_VOLTAGE_SETPOINT:
			vebusSystem.act.varValue = 0xFFFF;
			vebusGetVar(vebusSystem.master, vebusSystem.hub1.ramVarOffset + 1, getVoltageSetpointDone, &vebusSystem.act.varValue);
			break;
		case MK2_ACT_SOC:
			vebusGetVar(vebusSystem.master, VE_VAR_SOC_RAMVAR, getSocDone, &vebusSystem.act.varValue);
			break;
		case MK2_ACT_SET_IBAT_EXTRA_OR_SOC:
			if (vebusSystem.socControl.setExtraBatteryCurrent) {
				logI("mk2","broadcast I bat extra %d", vebusSystem.socControl.extraBatteryCurrent);
				vebusFrSendIBatExtra(sendIBatExtraDone, vebusSystem.socControl.extraBatteryCurrent);
			}
			else {
				logI("mk2","broadcast SOC raw %d", socToWrite);
				vebusFrSendSoc(sendIBatExtraDone, socToWrite);
			}
			break;
		default:
			veAssert(veFalse);
		}
	}
}

/*
 * Create a device list from the short ids.
 *
 * Note: the short id list is reset when it is read. So we
 * or here to keep a list of once known devices. Note, the mk2
 * can discover one vebus device per phase, so it takes about
 * 2 seconds before all devices are known.
 */
static void findNewDevice(void)
{
	un32 newId = vebusSystem.state.shortIds & ~vebusSystem.state.knownIds;
	un8 addr = 0;

	logI("task", "device ids %d", newId);

	while (newId)
	{
		if (newId & 1) {
			VebusDevice *dev;
			if (!(dev = vebusDeviceAlloc())) {
				logW("task", "device alloc failed %d", addr);
				return;
			}

			dev->addr = addr;
			dev->next = vebusSystem.devices;
			vebusSystem.devices = dev;
			vebusSystem.state.knownIds |= (1 << addr);
			logI("task", "found new device %d", dev->addr);
			mk2NewDeviceFound(dev);
		}

		addr++;
		newId >>= 1;
	}
}

static void mk2GetVebusStatusDone(void)
{
	// yes, this will get the version from the master again...
	vebusQueryDevice(vebusSystem.master, VE_UNITS_VSC, qeuryDone);
}

static void getNextFirmwareVersion(void)
{
	VebusDevice *dev = vebusSystem.init.dev;

	logI("task", "addr %hhu got version %u", dev->addr, dev->version);

	dev = dev->next;
	vebusSystem.init.dev = dev;

	if (dev == NULL) {
		mk2GetVebusState(&vebusSystem.state, mk2GetVebusStatusDone);
	} else {
		vebusGetTargetVersion(dev, getNextFirmwareVersion, &dev->version);
	}
}

/**
 * For a vebus workaround the firmware version of the slaves are needed.
 * So get them first.
 */
static void getAllFirmwareVersions(void)
{
	VebusDevice *dev;

	dev = vebusSystem.devices;
	vebusSystem.init.dev = dev;
	vebusGetTargetVersion(dev, getNextFirmwareVersion, &dev->version);
}


static void mk2GetShortIdsDone(void)
{
	logI("task", "short ids flags %X", vebusSystem.state.shortIds);

	/* Create device list */
	findNewDevice();

	if (vebusSystem.master) {
		/* If there is a vebus system, get the scaling from the master */
		valuesMasterDeviceFound();
		getAllFirmwareVersions();
	} else {
		/*
		* Otherwise power cycle the mk2; the mk2 only checks the system type
		* at startup and defaults to non vebus mode.
		*/
		logI("mk2", "no devices found, restarting");
		startupTimer = 20;
		mk2HoldInReset(veTrue);
	}
}

static void sendInitialPanelFrameDone(void)
{
	sendPanelFrame = PANEL_IDLE;
	vebusGetShortIds(mk2GetShortIdsDone);
}

static void waitForControlInit(void)
{
	/*
	 * Getting the short Ids and firmware versions can now fail, since they are obtained
	 * from the master and the firmware versions are asked to devices which should be there,
	 * but not necessarily are there. Or to put it otherwise, the startup will now always
	 * fail if a single device is having a problem. To make it worse, the vebus devices
	 * disable their leds when the panel detect is grounding, so the vebus error which
	 * is normally shown on the leds is no longer visible. To get them visible again, a
	 * panelFrame with a non off state must be send. So try to send that message when allowed.
	 */
	if (sendPanelFrame == PANEL_QUEUED) {
		logI("mk2", "sending panel frame");
		vebusSendPanelFrame(sendInitialPanelFrameDone);
		return;
	}

	vebusGetShortIds(mk2GetShortIdsDone);
}

static void bmsCheckDone(void)
{
	if (vebusSystem.bmsVersion != 0 && vebusSystem.bmsVersion != devRegStored.bmsVersion) {
		devRegStored.bmsVersion = vebusSystem.bmsVersion;
		storageSave();
	}

	vebusControlUpdate();
	delayCb = waitForControlInit;
	delayCount = 21;
}

static void dmcCheckDone(void)
{
	logI("done", "done");

	if (vebusSystem.dmcVersion != 0 && vebusSystem.dmcVersion != devRegStored.dmcVersion) {
		devRegStored.dmcVersion = vebusSystem.dmcVersion;
		storageSave();
	}

	/* if it is known that a BMS is present, don´t check it again */
	if (vebusSystem.bmsVersion) {
		bmsCheckDone();
		return;
	}

	vebusCheckBmsPresence(bmsCheckDone);
}

static void checkBlockModeDone(void)
{
	if (vebusSystem.state.flags & VEBUS_STATE_FLAG_MK2_BLOCKED) {
		valuesActionDoneCallback(MK2_ACT_STATUS);
		mk2GetVebusState(&vebusSystem.state, checkBlockModeDone);
		return;
	}

	/* if it is known that a DMC is present, don´t check it again */
	if (vebusSystem.dmcVersion) {
		dmcCheckDone();
		return;
	}

	vebusCheckDmcPresence(dmcCheckDone);
}

static void waitForMk2InitDone(void)
{
	mk2GetVebusState(&vebusSystem.state, checkBlockModeDone);
}

static void mk2SendSetStateDone(void)
{
	logI("mk2", "wait for mk2 init done (wait for L frame)");
	delayCb = waitForMk2InitDone;
	delayCount = 21;
}

/* 2> make sure panel frames are received */
static void mk2SendSetState(void)
{
	mk2.state.sFlags |= MK2_S_FLAG_AUTO_FORWARD_PANEL_FRAMES | MK2_S_FLAG_DO_NOT_SEND_STATE;
	mk2SetState(&mk2.state, mk2SendSetStateDone);
}

/* 2b< mk2/mk3 was there but outdated. */
static void mk2UpdateDone(Mk2Result result)
{
	VE_UNUSED(result);

	startupTimer = 20;
	mk2HoldInReset(veTrue);
	mk2EnsureVebusPower(veFalse);
}

/* 1< mk2 is there, set its state or update when outdated */
static void mk2GetVersionDone(void)
{
	un32 version;
	un32 product;

	version = mk2.version % 1000;
	product = mk2.version - version;

	switch (product) {
		/* MK2/MK3 returns the product ID in decimal. products.h defines them as 0x1170,0x1130,0x1131 */
#if CFG_WITH_MK3 == 1
		case (1170 * 1000):		/* VE_PROD_ID_MK3 */
			if (mk2.type != MK3) {
				logE("mk2", "Found MK3 expected type %d ", mk2.type);
				pltExit(20);
			}

			if ((version % 1000) != MK3_UPDATE_VERSION) {
				logI("mk2", "wrong MK3 firmware version %d, starting update to version %d", version, MK3_UPDATE_VERSION);
				mk3FirmwareUpgrade(mk2UpdateDone);
				return;
			}
			break;
#endif
		case (1130 * 1000): /* VE_PROD_ID_VEBUS_MK2_20MHZ_9915 */
		case (1131 * 1000):	/* VE_PROD_ID_VEBUS_MK2_DONGLE_20MHZ_9915 */
			if (mk2.type != MK2) {
				logE("mk2", "Found MK2 expected type %d ", mk2.type);
				pltExit(20);
			}
			if ((version % 1000) != mk2UpdateVersion) {
				logI("mk2", "wrong MK2 version %d, starting update to %d", version, mk2UpdateVersion);
				if (allowMk2Update)
					mk2FirmwareUpgrade(mk2UpdateDone);
				else
					logE("mk2", "MK2 needs updating but is not allowed");
				return;
			}
			break;
		default:
			logE("mk2", "Unknown product ID %d (/1000)", product);
			break;
	}

	if (mk2TakeOverCb) {
		mk2TakeOverCb(VACK_OK);
		mk2IsTakenOver = veTrue;
		mk2TakeOverCb = NULL;
		return;
	}

	/*
	 * Workaround for ccgx:
	 * - Also announce the presence of the device if it might not be on.
	 * Although there is limited information at least it can set the mode
	 * to ON / show that the device is off.
	 * - Likewise, also show the device if we are not allowed to control it.
	 * It is then visible in the menu and explained and 'redetect system' can
	 * be pressed.
	 */
	if (devRegStored.mode != SP_SWITCH_ON || vebusSystem.bmsVersion || vebusSystem.dmcVersion) {
		logI("task", "announcing device, since it is not on or out of our control");
		valuesExposeFakeDevice();
	}

	if (mk2.type == MK3) {
		if (pltUseHostMk2Io())
			/* For the MK3 we need to set the panel detect line first */
			mk3SetHardwareLines (mk2SendSetState, veTrue, veFalse);
		else
			/* A USB MK3 is powered through VE.Bus. We don't want it to act as panel */
			mk3SetHardwareLines (mk2SendSetState, veFalse, veFalse);
	}
	else
		mk2SendSetState();
}

static void getSettingOfDevice(VebusDevice *dev)
{
	VebusSettingInfo* scales;
	un8 n;

	scales = vebusGetSettingInfoOfDevice(dev);
	for (n = 0; n < VEBUS_SETTINGS_COUNT; n++)
		if (scales[n].scale)
			mk2ActGetRawSetting(dev, n);
}

static void currentSensorsPollDone(void)
{
	if (vebusAssistantsPoll(currentSensorsPollDone))
		return;

	logI("mk2", "init is done\n=========================================================");
	initDone = veTrue;
	valuesInitDoneCallback();

	logI("mk2", "getting ac config");
	vebusSystem.nextAction = MK2_ACT_AC_CONFIG;
	vebusGetAcInConfig(&vebusSystem.acInConfig, reqAcInConfigDone);

	{
		VebusDevice *dev = vebusSystem.devices;
		while (dev) {
			getSettingOfDevice(dev);
			dev = dev->next;
		}
	}
#if 0
	for (n = 0; n < VEBUS_RAMVAR_COUNT; n++)
		if (vebusScales.ramInfo[n].scale)
			mk2ActGetVar(vebusDevice, n);
#endif
}

void mk2ActSetVarDoneCb(struct VebusDeviceS *dev, un8 id)
{
	VE_UNUSED(id);

	valuesSetRamVarCallback(dev, id);
	queueNextAction(dev, -1);
}

/* Now all settings / variable info is available. */
static void qeuryDone(void)
{
#ifdef DBUS
	if (vebusSupportedIBatSocBroadcast(vebusSystem.master))
		valuesRegisterExtraBatteryCurrent();
#endif

	/* check if there are current sensors connected to the devices */
	vebusAssistantsPoll(currentSensorsPollDone);
}

/*
 * This is called early, there is nothing know yet about dev.
 * (besides its short address)
 */
void mk2NewDeviceFound(struct VebusDeviceS* dev)
{
	if (dev->addr == 0)
		vebusSystem.master = dev;
}

void taskInit(void)
{
#if !defined(TARGET_pc)
	allowMk2Update = veTrue;
#endif

	devRegInit();

	/* CAN bus specific values, perhaps just call invalidate? */
	vebusSystem.voltageSetpoint = 0xFFFF;
	vebusSystem.bmsVersion = devRegStored.bmsVersion;
	vebusSystem.dmcVersion = devRegStored.dmcVersion;
	/*
	 * note: since vebusError can cause init to fail, it should be send before init
	 * is done, on the otherhand, it takes some time to get it from the mk2, so 0xFF
	 * is abused as invalid value.
	 */
	vebusSystem.state.vebusError = 0xFF;

	/* Let the Multi calculate the Soc.
	 * Multi Soc calculation will be disabled after a new Soc is written */
	vebusSystem.socControl.extraBatteryCurrent = 0;
	vebusSystem.socControl.setExtraBatteryCurrent = veTrue;


#ifdef NMEA2K
	productInfoInit();
	msgRouterInit();
#endif

	mk2FhInit();
	mk2Init();
	mk2.type = mk2TypeInit;		/* Default is MK2, can be changed though console option */
	valuesInit();

	/*
	 * Setup the hardware, these are two distinct calls since the bpp3 does
	 * use the regular pc serial interface, but has hardware to control some
	 * mk2 io pins.
	 */
	mk2HalInit();
	mk2IoInit();

	mk2ActInit();
	vebusControlInit();
	vebusDeviceAllocInit();

	mk2BreakClear();
	mk2HoldInReset(veFalse); /* power the mk2 */
	if (powerControlAllowed())
		mk2EnsureVebusPower(veTrue); /* power vebus */
	mk2Begin();

	pltInterruptEnable();
}

void taskUpdate(void)
{
	valuesUpdate();

	mk2FhUpdate();
#ifdef NMEA2K
	msgRouterUpdate();
#endif
}

void taskTick(void)
{
#ifdef NMEA2K
	msgRouterTick();
#endif

	if (mk2UpdateBusy()) {
		mk2FirmwareTick();
		return;
	}

	mk2FhTick();

#if CFG_WITH_MK3 == 1
	if (mk3FirmwareUpgradeBusy()) {
		mk3FirmwareUpgradeTick();
		return;
	}
#endif

	vebusControlTick();
	storageTick();

	if (startupTimer == 20) {
		invalidate();

		/* Power Ve.Bus but only if it needs to be on and no BMS etc control it */
		if (powerControlAllowed())
			mk2EnsureVebusPower(veTrue);
	}

	/*
	 * Get the mk2 out of reset before the startupTimer expires, so it has
	 * some time to startup.
	 */
	if (startupTimer == 5)
		mk2HoldInReset(veFalse);

	/* begin the startup procedure */
	if (startupTimer != 0 && --startupTimer==0)
		mk2Begin();

	valuesTick();

	if (kwhTimer != 0 && --kwhTimer == 0)
		getKwhValues = veTrue;

	if (delayCount && --delayCount == 0)
		delayCb();
}

/* callback from mk2_action after a variable is obtained */
void mk2ActObtainedRawVarCb(struct VebusDeviceS* dev, un8 id, un16 value)
{
	dev->vars[id] = value;
	valuesRamVarObtainedCallback(dev, id);
	queueNextAction(dev, MK2_ACT_DEV_GET_RAM_VAR);
}

veBool vebusIsValidSetting(struct VebusDeviceS* dev, un8 id)
{
	return dev->settingValidFlags[id / 8] & (1 << (id & 7));
}

void vebusSetSettingValid(struct VebusDeviceS* dev, un8 id)
{
	dev->settingValidFlags[id / 8] |= (1 << (id & 7));
}

void vebusSetSettingInvalid(struct VebusDeviceS* dev, un8 id)
{
	dev->settingValidFlags[id / 8] &= ~(1 << (id & 7));
}

/* callback from mk2_action after a setting is obtained */
void mk2ActObtainedRawSettingCb(struct VebusDeviceS* dev, un8 id, un16 value)
{
	veBool changed = dev->settings[id] != value;

	logI("mk2", "got setting %d", id);
	dev->settings[id] = value;
	vebusSetSettingValid(dev, id);
	if (changed)
		valuesSettingChangedCallback(dev, id);
	queueNextAction(dev, MK2_ACT_DEV_GET_SETTING);
}

void mk2ActDoneCb(struct VebusDeviceS* dev, Mk2Actions action)
{
	logI("task", "mk2ActDoneCb");
	queueNextAction(dev, action);
}

un16 mk2ActSetRawSettingCb(struct VebusDeviceS* dev, un8 id)
{
	return dev->settings[id];
}

void mk2ActSetSettingDoneCb(struct VebusDeviceS* dev, un8 id)
{
	valuesSetSettingDoneCallback(dev, id);
	mk2ActDoneCb(dev, MK2_ACT_DEV_SET_SETTING);
}

/* 1> first get the version of the mk2 */
static void mk2Begin(void)
{
	mk2.version = 0;

#if CFG_WITH_CONTROL
	/* init to stored version, to prevent unnecesarry flash writes */
	vebusSetModeVariable(devRegStored.mode);
#endif

	mk2GetVersion(&mk2.version, mk2GetVersionDone);
}

/** No response is received after several tries. */
void mk2NoResponseCallback(void)
{
	invalidate();

	logI("mk2", "timeout");

#if CFG_WITH_EXIT_ON_NO_CONNECTION
	if (pltExitOnNoConnection())
		pltExit(0);
#endif

	if (devRegStored.mode == SP_SWITCH_OFF) {
		if (mk2.type == MK2) {
			/*
			 * The multi is off and thus the mk2 unreachable (when powered from the
			 * vebus side). So it is fine. Another device can turn the multi on though,
			 * so check later again if the mk2/multi isn't started by another device.
			 */
			startupTimer = 20;
			mk2HoldInReset(veTrue);
		}
		else
			startupTimer = 20;
		/*
		 * we can get stuck in this loop if there is no valid firmware in the mk2, since
		 * we never get a valid version frame, which is what we are waiting for. We also
		 * don't want to power the vebus when is supposed to be off. So for now do announce
		 * a fake device so at least on the ccgx e.g. a menu is presented (and user can
		 * get us out of this loop by turning the device on).
		 */
		logI("task", "announcing off device since mk2 did not respond to version");
		valuesExposeFakeDevice();

		return;
	}

	/* if the mk2/mk3 didn't respond, try to update it.. */
	if (!mk2.version) {
#if CFG_WITH_MK3 == 1
		if (mk2.type == MK3) {
			mk3FirmwareUpgrade(mk2UpdateDone);
			return;
		}
#else
		if ((allowMk2Update) && (mk2.type == MK2) {
			mk2FirmwareUpgrade(mk2UpdateDone);
			return;
		}
		logW("mk2", "not allowed to update the mk2!");
#endif
	}

	startupTimer = 20;
	mk2HoldInReset(veTrue);
	mk2EnsureVebusPower(veFalse);
}

/* If the control code want to send a panel frame, it is queued in between the normal communication */
void vebusQueuePanelFrameCallback(void)
{
#if CFG_WITH_CONTROL
	if (sendPanelFrame == PANEL_IDLE)
		sendPanelFrame = PANEL_QUEUED;
#endif
}

/* Writing SOC is part of some control loops, so it goes in between normal communication */
void mk2WriteSoc(un16 rawValue)
{
	if (sendSoc == PANEL_IDLE) {
		socToWrite = rawValue;
		vebusSystem.socControl.setExtraBatteryCurrent = veFalse;
		sendSoc = PANEL_QUEUED;
	}
}

void mk2TakeOverRequest(mk2TakeOverRequestCb *cb)
{
	veAssert(cb);

	if (mk2TakeOverCb) {
		cb(VACK_ERR_BUSY | 1);
		return;
	}
	if (mk2IsTakenOver) {
		cb(VACK_ERR_BUSY | 2);
		return;
	}
	if (vebusSystem.state.flags & VEBUS_STATE_FLAG_MK2_BLOCKED) {
		cb(VACK_ERR_BUSY | 3);
		return;
	}
	if (mk2UpdateBusy()) {
		cb(VACK_ERR_BUSY | 4);
		return;
	}

	if (!mk2.version) {
		cb(VACK_ERR_INITIALIZING);
		return;
	}

	mk2TakeOverCb = cb;

	invalidate();
	startupTimer = 20;

	if (mk2.type == MK2)
		mk2HoldInReset(veTrue);	/* The MK2 could be in VE.9bit mode. Unsupported by MK3 */

	mk2EnsureVebusPower(veFalse);
}

void mk2SetProductType(Mk2Type type)
{
	mk2TypeInit = type;
}
