#include <stdlib.h>

#include <velib/velib_inc_mk2.h>

#include <velib/base/base.h>
#include <velib/base/ve_string.h>
#include <velib/io/ve_remote_device_item.h>
#include <velib/mk2/mk2_info.h>
#include <velib/mk2/mk2_remote_device.h>
#include <velib/mk2/hal.h>
#include <velib/mk2/values_float.h>
#include <velib/mk2/vebus_items.h>
#include <velib/platform/console.h>
#include <velib/platform/plt.h>
#include <velib/types/variant_print.h>
#include <velib/types/ve_item.h>
#include <velib/types/ve_str.h>
#include <velib/types/ve_values.h>
#include <velib/utils/ve_logger.h>
#include <velib/utils/ve_item_utils.h>

#include <bus_values.h>
#include <dev_reg.h>
#include <storage.h>
#include <task.h>
#include <version.h>

#define HUB4_DISABLE_CHARGE_FLAG		0x0001
#define HUB4_DISABLE_FEEDBACK_FLAG		0x0002
#define HUB4_SUSTAIN_FLAG				0x0100
#define HUB4_LOW_SOC_FLAG				0x0200

/** The root of the tree to locate values and format them etc */
static VeItem root;

/**
 * The vebus system found on the bus containing the values also accessible
 * from the tree.
 */
static VeBusItems veBusItems;

static VeItem *state;
static VeItem *currentLimit;
static VeItem *modeItem;
static VeItem *redetectSystem;
static VeItem *resetSystem;
static VeItem *connected;
static VeItem *productId;
static VeItem *firmwareVersion;
static VeItem *maxChargeCurrent;
static VeItem *extraBatteryCurrent = NULL;
static veBool exported;
static char const version[] = VERSION_STR;

/* More specials since a common assistant interface is still lacking.. */
static VeItem *hub1VoltageSetPoint;

/* hub4 control */
typedef struct {
	VeItem *setPoint;
	VeItem *disableCharge;
	VeItem *disableFeedback;
} Hub4Assistants;

typedef struct {
	VeItem *compatSetPoint;
	VeItem *disableCharge;
	VeItem *disableFeedback;
	VeItem *sustain;
	VeItem *lowSoc;
} Hub4Interface;

static Hub4Assistants *hub4assistants[3];
static Hub4Interface *hub4Interface;

static VeItemRemoteDev mk2ItemDevice;

char const *pltProgramVersion(void)
{
	return version;
}

void valuesUpdate(void)
{
	static int count;

	valuesDbusUpdate();

	if (++count > 100) {
		count = 0;

		if (hub4Interface && vebusSystem.master)
			mk2ActGetAssistantVar(vebusSystem.master, vebusItemGetVarNr(hub4Interface->sustain));
	}
}

/** 50ms tick to invalidate items if there are not received in time */
void valuesTick(void)
{
#if CFG_WITH_VE_REMOTE_DEV
	mk2RemoteDevTick();
#endif
	veItemTick(&root);
}

void vebusModeChanged(VebusSwitchPos mode)
{
	VeVariant tmp;

	veItemOwnerSet(modeItem, veVariantUn32(&tmp, mode));

	if (devRegStored.mode != mode) {
		devRegStored.mode = mode;
		storageSave();
	}
}

void vebusCurrentLimitChanged(un8 acin, un16 limit)
{
	VeVariant tmp;
	VE_UNUSED(acin);

	veItemOwnerSet(currentLimit, veVariantFloat(&tmp, (float) limit / 10));
}

#if CFG_WITH_CONTROL
static veBool setCurrentLimit(un8 n, VeVariant *var)
{
	un16 limit;

	/* convert to int in 100mA for vebus */
	VeVariant conv = *var;
	veVariantToFloat(&conv);
	conv.value.Float *= 10;
	veVariantToN32(&conv);
	limit = (un16) conv.value.UN32;

	/* If the current limit changed -> store it */
	if (devRegStored.currentLimit[n] != limit) {
		devRegStored.currentLimit[n] = limit;
		storageSave();
	}

	/* If this is the active input attempt to set it */
	if (vebusActiveIn(&vebusSystem.acInConfig) == n)
		vebusControlDevice(NULL, SP_UNKNOWN);

	return veTrue;
}
#else
static veBool setCurrentLimit(un8 n, VeVariant *var)
{
	VE_UNUSED(n);
	VE_UNUSED(var);

	return veFalse;
}
#endif

/* stub to forward current limit requests for the active input */
static veBool veItemSetCurrentLimit(struct VeItem *item, void *ctx, VeVariant *variant)
{
	un8 input;
	VE_UNUSED(item);
	VE_UNUSED(ctx);

	input = vebusActiveIn(&vebusSystem.acInConfig);
	return setCurrentLimit(input, variant);
}

/* stub to forward current limit requests for the active input */
static veBool veItemRedetectSystem(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VeVariant tmp;
	VE_UNUSED(ctx);
	VE_UNUSED(variant);

	/* reset the stored bms/dmc version */
	devRegStored.bmsVersion = 0;
	devRegStored.dmcVersion = 0;
	storageSave();
	mk2InfoClearCache();

	/* Change value for onChangeEvents */
	veItemOwnerSet(item, veVariantUn32(&tmp, 1));

	/* Trigger a complete reinit */
	mk2NoResponseCallback();

	return veTrue;
}

/* stub to forward current SOC */
static veBool veItemSetSoc(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VebusDevice *dev = (VebusDevice *) ctx;
	VE_UNUSED(item);

	if (!vebusSupportedRamVar(dev, VE_VAR_SOC_RAMVAR))
		return veFalse;

	/*
	 * note: there are to conversion here combined
	 * First of all vebus uses 0..1 for SOC.
	 * This is normally encoded with scale 200, so the raw values are 0..200.
	 * The second parts is obtained at runtime though and can change.
	 */
	veVariantToFloat(variant);
	variant->value.Float *= 0.01 / mk2UnpackRamFloat(1, dev, VE_VAR_SOC_RAMVAR);
	veVariantToN32(variant);

	mk2WriteSoc(variant->value.UN32);

	return veTrue;
}

#if CFG_WITH_CONTROL
/* stub to forward state change request. */
static veBool veItemSetMode(struct VeItem *item, void *ctx, VeVariant *variant)
{
	un32 mode;
	VE_UNUSED(item);
	VE_UNUSED(ctx);

	veVariantToN32(variant);
	mode = variant->value.UN32;
	if (mode < SP_FIRST || mode > SP_LAST)
		return veFalse;

	/* Try to get control over the multi */
	vebusControlDevice(NULL, mode);

	return veTrue;
}
#else
static veBool veItemSetMode(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VE_UNUSED(item);
	VE_UNUSED(ctx);
	VE_UNUSED(variant);

	return veFalse;
}
#endif

static veBool veItemSetMaxChargeCurrent(struct VeItem *item, void *ctx, VeVariant *variant)
{
	un16 value;
	VE_UNUSED(item);
	VE_UNUSED(ctx);

	if (!vebusSystem.master || !vebusSystem.ac[0].phaseCount)
		return veFalse;

	/* XXX: needs to be the number of devices... */
	veVariantToFloat(variant);
	variant->value.Float /= vebusSystem.ac[0].phaseCount;
	value = mk2PackSettingFloat(variant->value.Float, vebusSystem.master, VE_I_BAT_BULK, VE_UNITS_VSC);
	vebusSystem.master->settings[VE_I_BAT_BULK] = value;
	mk2ActSetRawSetting(vebusSystem.master, VE_I_BAT_BULK);

	return veTrue;
}

static veBool veItemResetSystem(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VE_UNUSED(item);
	VE_UNUSED(ctx);

	mk2ActMark(MK2_ACT_SYSTEM_RESET);
	veItemOwnerSet(resetSystem, variant);

	return veTrue;

}

size_t veStateString(VeVariant *var, void const *ctx, char *buf, size_t len)
{
	VE_UNUSED(ctx);

	if (!veVariantIsValid(var)) {
		buf[0] = 0;
		return 0;
	}

	return vebusNmea2000StateText(var->value.UN32, buf, len);
}

static size_t hexString(VeVariant *var, void const *ctx, char *buf, size_t len)
{
	VE_UNUSED(ctx);

	if (!veVariantIsValid(var)) {
		buf[0] = 0;
		return 0;
	}

	return ve_snprintf(buf, len, "%X", var->value.UN32);
}

/*
 * Create a vebus system object and the corresponding item tree.
 *
 * The tree with items can be used to navigate through the values in
 * vebus system, format / set / get them etc.
 */
VeBusItems* createVebusSystem(VeItem* root)
{
	VeVariant tmp;
	VeItem *item;

	vebusItemsCreateSystem(root, &veBusItems);

	state = veItemGetOrCreateUid(root, "State");
	veItemSetFmt(state, veStateString, NULL);

	modeItem = veItemGetOrCreateUid(root, "Mode");
	currentLimit = veItemGetOrCreateUid(root, "Ac/ActiveIn/CurrentLimit");

	veItemCreateBasic(root, "DeviceInstance", mk2Instance(&tmp));
	veItemCreateBasic(root, "Mgmt/ProcessName", veVariantStr(&tmp, pltProgramName()));
	veItemCreateBasic(root, "Mgmt/ProcessVersion", veVariantStr(&tmp, version));
	veItemCreateBasic(root, "Mgmt/Connection", veVariantStr(&tmp, "VE.Bus"));

	connected = veItemCreateBasic(root, "Connected", veVariantUn8(&tmp, 0));
	productId = veItemGetOrCreateUid(root, "ProductId");
	veItemSetFmt(productId, hexString, 0);
	firmwareVersion = veItemGetOrCreateUid(root, "FirmwareVersion");
	veItemSetFmt(firmwareVersion, hexString, 0);

	redetectSystem = veItemGetOrCreateUid(root, "RedetectSystem");
	veItemSetSetter(redetectSystem, veItemRedetectSystem, NULL);
	veItemOwnerSet(redetectSystem, veVariantUn32(&tmp, 0));

	item = veItemGetOrCreateUid(root, "Soc");
	veItemSetSetter(item, veItemSetSoc, &veBusItems);

	resetSystem = veItemGetOrCreateUid(root, "SystemReset");
	veItemSetSetter(resetSystem, veItemResetSystem, NULL);

	veItemSetSetter(currentLimit, veItemSetCurrentLimit, vebusSystem.master);
	veItemSetSetter(modeItem, veItemSetMode, &veBusItems);
	veItemSetMin(currentLimit, veVariantFloat(&tmp, 0));
	veItemSetMax(currentLimit, veVariantFloat(&tmp, 40));

	maxChargeCurrent = veItemByUid(root, "Dc/0/MaxChargeCurrent");
	veItemSetSetter(maxChargeCurrent, veItemSetMaxChargeCurrent, NULL);

#if CFG_WITH_VE_REMOTE_DEV
	item = veItemGetOrCreateUid(root, "Interfaces/Mk2/Tunnel");
	veItemRemoteDevInit(&mk2ItemDevice, item, VE_REMOTE_DEV_ROLE_SERVICE);
	mk2RemoteDevInit(&mk2ItemDevice.remoteDev);
#endif

	return &veBusItems;
}

void valuesInit(void)
{
	char const *device = pltGetSerialDevice();
	VeStr s;
	VeItem *itemTree = veValueTree();

	veStrNew(&s, 100, 100);
	veStrAdd(&s, "com.victronenergy.vebus");
	if (strstr(device, "/dev/tty") == device)
		veStrAddFormat(&s, ".%s", device + strlen("/dev/"));

	createVebusSystem(&root);
	veItemAddChildByUid(itemTree, s.data, &root);

	valuesDbusInit(&root, s.data);

	veStrFree(&s);
};

/* Called after a set or a get of a ramvar */
void valuesRamVarMightHaveChanged(struct VebusDeviceS *dev, un8 id)
{
	VeVariant var;
	un8 phase = dev->addr % 3;

	if (hub4assistants[phase]) {
		if (id == vebusItemGetVarNr(hub4assistants[phase]->setPoint)) {
			veVariantSn32(&var, (sn16) dev->vars[id] * -1 * vebusSystem.ac[0].inverterFactor);
			veItemOwnerSet(hub4assistants[phase]->setPoint, &var);
			veItemOwnerSet(hub4Interface->compatSetPoint, &var);
		}

		if (id == vebusItemGetVarNr(hub4assistants[phase]->disableCharge)) {
			veItemOwnerSet(hub4assistants[phase]->disableCharge, veVariantSn32(&var, (dev->vars[id] & HUB4_DISABLE_CHARGE_FLAG) != 0));
			veItemOwnerSet(hub4assistants[phase]->disableFeedback, veVariantSn32(&var, (dev->vars[id] & HUB4_DISABLE_FEEDBACK_FLAG) != 0));

			if (dev->addr == 0) {
				veItemOwnerSet(hub4Interface->disableCharge, veVariantSn32(&var, (dev->vars[id] & HUB4_DISABLE_CHARGE_FLAG) != 0));
				veItemOwnerSet(hub4Interface->disableFeedback, veVariantSn32(&var, (dev->vars[id] & HUB4_DISABLE_FEEDBACK_FLAG) != 0));
				veItemOwnerSet(hub4Interface->lowSoc, veVariantSn32(&var, (dev->vars[id] & HUB4_LOW_SOC_FLAG) != 0));
				veItemOwnerSet(hub4Interface->sustain, veVariantSn32(&var, (dev->vars[id] & HUB4_SUSTAIN_FLAG) != 0));
			}
		}
	}
}

/* Hooks from task.c to do application specific things with the obtained values */

void valuesRamVarObtainedCallback(struct VebusDeviceS* dev, un8 id)
{
	vebusItemsRamVarObtained(&veBusItems, dev, id);
	valuesRamVarMightHaveChanged(dev, id);
}

void valuesActionDoneCallback(Mk2Actions done)
{
	VeVariant tmp;

	if (!exported)
		return;

	if (vebusSystem.state.flags & VEBUS_STATE_FLAG_MK2_BLOCKED) {
		veItemSet(state, veVariantUn32(&tmp, N2K_CONVERTER_STATE_BLOCKED));
		return;
	}

	vebusItemsActionDone(&veBusItems, vebusSystem.master, done);

	switch ((int) done) {
	case MK2_ACT_SYSTEM_RESET:
		veItemOwnerSet(resetSystem, veVariantUn32(&tmp, 0));
		break;

	case MK2_ACT_CONV_STATE:
		veItemSet(state, veVariantUn32(&tmp, vebusN2kStateConverter()));
		break;

	case MK2_ACT_AC_VALUES:
	{
		double max;

		max = vebusGetSettingMaximum(vebusSystem.master, VE_I_BAT_BULK, VE_UNITS_VSC) * vebusSystem.ac[0].phaseCount;
		veItemSetMax(maxChargeCurrent, veVariantFloat(&tmp, max));
		break;
	}
	case MK2_ACT_AC_CONFIG:
		veItemSetMax(currentLimit, veVariantFloat(&tmp, vebusSystem.acInConfig.max / 10.0f));
		break;
	case 99: // FIXME, = voltage setpoint from hub1 obtained
		if (hub1VoltageSetPoint != NULL) {
			if (vebusSystem.voltageSetpoint == 0xFFFF)
				veItemInvalidate(hub1VoltageSetPoint);
			else
				veItemOwnerSet(hub1VoltageSetPoint, veVariantFloat(&tmp, vebusSystem.voltageSetpoint / 100.0f));
		}
		break;

	default:
		;
	}
}

/* stub to forward current limit requests for the active input */
static veBool veItemSetHub4Power(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VebusDevice *dev = vebusItemGetDev(item);
	un8 id = vebusItemGetVarNr(item) & 0xFF;
	un8 phase = dev->addr % 3;
	VE_UNUSED(ctx);

	if (!veVariantIsValid(variant)) {
		return veFalse;
	}

	veVariantToN32(variant);

	if (item == hub4assistants[phase]->setPoint) {
		Mk2RawAcPhase *acPhaseInfo = &vebusSystem.ac[phase];

		variant->value.SN32 *= -1;
		if (acPhaseInfo->inverterFactor == 0)
			return veFalse;
		variant->value.SN32 /= acPhaseInfo->inverterFactor;

		/* Clamp the value to fit in a RAMVAR */
		if (variant->value.SN32 > 0x7FFE) {
			variant->value.SN32 = 0x7FFE;
			logI("hub4", "clamping setpoint to prevent overflow");
		} else if (variant->value.SN32 < -0x7FFE) {
			variant->value.SN32 = -0x7FFE;
			logI("hub4", "clamping setpoint to prevent underflow");
		}
	}

	dev->vars[id] = (sn16) variant->value.SN32;
	logI("hub4", "request to set %d %hhu to %hd", dev->addr, id, dev->vars[id]);

	mk2ActSetAssistantVar(dev, id);

	return veTrue;
}

static veBool veItemSetHub4Flag(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VebusDevice *dev = vebusItemGetDev(item);
	un8 id = vebusItemGetVarNr(item) & 0xFF;
	un8 phase = dev->addr % 3;
	VE_UNUSED(ctx);

	if (!veVariantIsValid(variant)) {
		return veFalse;
	}

	veVariantToN32(variant);

	if (item == hub4assistants[phase]->disableCharge) {
		dev->vars[id] &= ~HUB4_DISABLE_CHARGE_FLAG;
		if (variant->value.UN32)
			dev->vars[id] |= HUB4_DISABLE_CHARGE_FLAG;
	} else	if (item == hub4assistants[phase]->disableFeedback) {
		dev->vars[id] &= ~HUB4_DISABLE_FEEDBACK_FLAG;
		if (variant->value.UN32)
			dev->vars[id] |= HUB4_DISABLE_FEEDBACK_FLAG;
	}

	logI("hub4", "set %d %hhu to %hX\n", dev->addr, id, dev->vars[id]);

	mk2ActSetAssistantVar(dev, id);

	return veTrue;
}

void valuesInvalidate(void)
{
	VeVariant v;

	vebusItemsInvalidate(&veBusItems, &root);

	if (mk2.state.mode != SP_SWITCH_OFF)
		veItemInvalidate(state);

	if (hub1VoltageSetPoint != NULL)
		veItemInvalidate(hub1VoltageSetPoint);
	veItemSet(connected, veVariantUn8(&v, 0));
}

// Either because it is set or requested it
void valuesSettingChanged(struct VebusDeviceS* dev, un8 id)
{
	VeVariant v;

	if (dev == vebusSystem.master && id == VE_I_BAT_BULK) {
		float max = mk2UnpackSettingFloat(dev->settings[VE_I_BAT_BULK], dev, VE_I_BAT_BULK, VE_UNITS_VSC);
		max *= vebusSystem.ac[0].phaseCount;
		veItemOwnerSet(maxChargeCurrent, veVariantFloat(&v, max));
	}
}

void valuesSettingChangedCallback(struct VebusDeviceS* dev, un8 id)
{
	vebusItemsSettingChanged(&veBusItems, dev, id);
	valuesSettingChanged(dev, id);
}

void valuesSetSettingDoneCallback(struct VebusDeviceS* dev, un8 id)
{
	vebusItemsSettingChanged(&veBusItems, dev, id);
	valuesSettingChanged(dev, id);
}

/* not short on memory, just keep all ramvars around */
un16 mk2ActSetVarCb(struct VebusDeviceS* dev, un8 id)
{
	return dev->vars[id];
}

void valuesSetRamVarCallback(struct VebusDeviceS *dev, un8 id)
{
	valuesRamVarMightHaveChanged(dev, id);
}

// NOTE: hub4 requires 26xx304 firmware!
static void registerHub4Assistant(char const *prefix, VebusDevice *dev, un16 ramVarOffset)
{
	VeItem *hub4Item;
	VeVariant v;
	un16 id;
	un8 phase = dev->addr % 3;

	if (dev->addr >= 3) {
		logW("", "hub4 assistant found in slave!");
		return;
	}

	if (hub4assistants[phase] != NULL) {
		logW("", "multiple hub4 assistant found in master!");
		return;
	}

	hub4assistants[phase] = malloc(sizeof(Hub4Assistants));
	if (hub4assistants[phase] == NULL) {
		logE("", "out of memory!");
		return;
	}

	logI("", "hub4 assistant found");

	hub4Item = veItemGetOrCreateUid(&root, prefix);
	id = ramVarOffset + 1;
	hub4assistants[phase]->setPoint = vebusItemCreateQuantity(hub4Item, dev, id, "AcPowerSetpoint", veVariantSn32(&v, 0), &veUnitWatt);
	veItemSetSetter(hub4assistants[phase]->setPoint, veItemSetHub4Power, NULL);
	mk2ActGetAssistantVar(dev, id);

	// Split the bits
	id = ramVarOffset + 2;
	hub4assistants[phase]->disableCharge = vebusItemCreateQuantity(hub4Item, dev, id, "DisableCharge", veVariantInvalidType(&v, VE_UN32), &veUnitNone);
	veItemSetSetter(hub4assistants[phase]->disableCharge, veItemSetHub4Flag, dev);
	hub4assistants[phase]->disableFeedback = vebusItemCreateQuantity(hub4Item, dev, id, "DisableFeedback", veVariantInvalidType(&v, VE_UN32), &veUnitNone);
	veItemSetSetter(hub4assistants[phase]->disableFeedback, veItemSetHub4Flag, dev);
	mk2ActGetAssistantVar(dev, id);
}

static veBool forwardHub4Setpoint(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VE_UNUSED(item);
	VE_UNUSED(ctx);
	return veItemSet(hub4assistants[0]->setPoint, variant);
}

static veBool forwardHub4ToAll(struct VeItem *item, void *ctx, VeVariant *variant)
{
	size_t n;
	VE_UNUSED(ctx);

	for (n = 0; n < ARRAY_LENGTH(hub4assistants); n++) {
		VeItem *tg;
		Hub4Assistants *hub4 = hub4assistants[n];

		if (!hub4)
			continue;

		if (item == hub4Interface->disableCharge)
			tg = hub4->disableCharge;
		else if (item == hub4Interface->disableFeedback)
			tg = hub4->disableFeedback;
		else
			return veFalse;
		veItemSet(tg, variant);
	}

	return veTrue;
}

static void registerHub4Interface(un16 ramVarOffset)
{
	VeItem *hub4Item;
	VeVariant v;
	un16 id;
	VebusDevice *dev = vebusSystem.master;

	if (hub4Interface)
		return;

	hub4Interface = malloc(sizeof(Hub4Interface));

	/* for compatibility, shoud be removed */
	hub4Item = veItemGetOrCreateUid(&root, "Hub4");
	id = ramVarOffset + 1;
	hub4Interface->compatSetPoint = vebusItemCreateQuantity(hub4Item, dev, id, "AcPowerSetpoint", veVariantSn32(&v, 0), &veUnitWatt);
	veItemSetSetter(hub4Interface->compatSetPoint, forwardHub4Setpoint, NULL);

	/* send flags to all of them */
	id = ramVarOffset + 2;
	hub4Interface->disableCharge = vebusItemCreateQuantity(hub4Item, dev, id, "DisableCharge", veVariantInvalidType(&v, VE_UN32), &veUnitNone);
	veItemSetSetter(hub4Interface->disableCharge, forwardHub4ToAll, dev);
	hub4Interface->disableFeedback = vebusItemCreateQuantity(hub4Item, dev, id, "DisableFeedback", veVariantInvalidType(&v, VE_UN32), &veUnitNone);
	veItemSetSetter(hub4Interface->disableFeedback, forwardHub4ToAll, dev);

	hub4Interface->lowSoc = vebusItemCreateQuantity(hub4Item, dev, id, "LowSoc", veVariantInvalidType(&v, VE_UN32), &veUnitNone);
	hub4Interface->sustain = vebusItemCreateQuantity(hub4Item, dev, id, "Sustain", veVariantInvalidType(&v, VE_UN32), &veUnitNone);
}

void vebusAssistantFoundCallback(VebusDevice *dev, un8 assistentId, un8 ramVarOffset, un8 ramvarCount)
{
	VE_UNUSED(ramvarCount);

	switch (assistentId)
	{
	case 2: // FIXME
		{
			if (hub1VoltageSetPoint != NULL)
				break;
			hub1VoltageSetPoint = veItemGetOrCreateUid(&root, "Hub1/ChargeVoltage");
		}
		break;
	case 3: // FIXME
		{
			un8 phase = dev->addr % 3;
			VeStr s;

			veStrNewFormat(&s, "Hub4/L%d", phase + 1);
			registerHub4Assistant(veStrCStr(&s), dev, ramVarOffset);
			veStrFree(&s);

			if (dev->addr == 0)
				registerHub4Interface(ramVarOffset);
		}
		break;
	}
}

/*
 * Note: there is very limited information about the system at this point in time.
 * The mk2 is found and vebus device short ids are known, alarms and leds are obtained.
 * Version of the vebus is unknown yet though, no scaling etc.
 */
static void updateEarlyInfo(void)
{
	vebusItemUpdate(&veBusItems.VE_SYS_MK2_VERSION);
	vebusItemUpdate(&veBusItems.VE_SYS_MK2_PRODUCT_ID);
	vebusItemUpdate(&veBusItems.VE_SYS_MK2_PRODUCT_NAME);
	vebusItemUpdate(&veBusItems.VE_SYS_SHORT_IDS);
	vebusItemsActionDone(&veBusItems, NULL, MK2_ACT_STATUS);
}

/*
 * On a ccgx there is a mk2 onboard. Since the mk2 will always
 * be found, wait for a master device to be found before exporting
 * to the dbus.
 */
void valuesMasterDeviceFound(void)
{
	if (exported) {
		updateEarlyInfo();
		return;
	}

	exported = veTrue;
	valuesDbusPublish();
	updateEarlyInfo();
}

/**
 * Hackery for the color control.
 * When a vebus system is off or out of our control, a menu is hacked in
 * to allow a user to change the mode or at least see it. This can be called
 * anywhere during startup, so information is not complete. It is also called
 * when polling for a mk2 version fails since there is no good reason for that
 * to fail on a ccgx.
 */
void valuesExposeFakeDevice(void)
{
	VeVariant v;

	/* Export fake vebus device */
	valuesMasterDeviceFound();

	/*
	 * The control state machine might not be active. Bypass it so it only depends on
	 * known / stored information.
	 */
	if (vebusCanBeControlled())
		veItemOwnerSet(&veBusItems.VE_SYS_MODE_IS_ADJUSTABLE, veVariantUn32(&v, 1));
	else
		veItemOwnerSet(&veBusItems.VE_SYS_MODE_IS_ADJUSTABLE, veVariantUn32(&v, 0));
}

void valuesInitDoneCallback(void)
{
	VeVariant tmp;

	vebusItemsInitDone(&veBusItems, &root, vebusSystem.master);
	veItemSet(connected, veVariantUn8(&tmp, 1));

	/* enable redetect */
	veItemOwnerSet(redetectSystem, veVariantUn32(&tmp, 0));
	veItemOwnerSet(productId, veVariantUn32(&tmp, mk2VersionToProductBcd(vebusSystem.master->version.number)));
	veItemOwnerSet(firmwareVersion, veVariantUn32(&tmp, mk2VersionToBcd(vebusSystem.master->version.number)));
}

static veBool veItemSetExtraBatteryCurrent(struct VeItem *item, void *ctx, VeVariant *var)
{
	VeVariant conv = *var;
	sn32 current;

	VE_UNUSED(ctx);

	/* convert to int in 100mA for vebus */
	veVariantToFloat(&conv);
	conv.value.Float *= 10;
	veVariantToN32(&conv);
	current = conv.value.UN32;
	logI("extraIBat", "Writing extra battery current value %d",current);

	/* Clamp the value to fit in a RAMVAR */
	if (current > 0x7FFFFF) {
		current = 0x7FFFFF;
		logI("extraIBat", "clamping current to prevent overflow");
	} else if (current < -0x7FFFFE) {
		current = -0x7FFFFE;
		logI("extraIBat", "clamping current to prevent underflow");
	}

	veItemOwnerSet(item, veVariantFloat(&conv, (float) current / 10));

	vebusSystem.socControl.extraBatteryCurrent = current;
	/* Schedule MK2_ACT_SET_IBAT_EXTRA_OR_SOC action. */
	vebusSystem.socControl.setExtraBatteryCurrent = veTrue;


	return veTrue;
}

void valuesRegisterExtraBatteryCurrent(void)
{
	VebusDevice *dev = vebusSystem.master;
	VeVariant v;

	if (extraBatteryCurrent == NULL) {
		extraBatteryCurrent = malloc(sizeof(VeItem));
		if (extraBatteryCurrent == NULL) {
			logE("extraIBat", "out of memory!");
			return;
		}
	}

	logI("extraIBat", "Register item ExtraBatteryCurrent");

	/* since this is no RAMVAR we don't need a var item nr. so set to 0
	 * Init variant value to 0 */
	extraBatteryCurrent = vebusItemCreateQuantity(&root, dev, 0, "ExtraBatteryCurrent", veVariantFloat(&v, 0), &veUnitAmps1Dec);

	veItemSetSetter(extraBatteryCurrent, veItemSetExtraBatteryCurrent, NULL);
}
